from __future__ import annotations  # noqa: F404

from typing import Dict, Hashable, List

import pandas as pd
from natsort import natsorted
from pygef.common import Location

from pypilecore.results.result_definitions import (
    CPTGroupResultDefinitions,
    CPTResultDefinitions,
)
from pypilecore.results.typing import MultiCPTBearingResults


class CasesMultiCPTBearingResults:
    """
    Container class for the results of multiple cases of MultiCPTBearingResults.
    It expects that all MultiCPTBearingResults objects have the same pile tip levels and same test ids.
    """

    def __init__(
        self,
        results_per_case: Dict[Hashable, MultiCPTBearingResults],
        cpt_locations: Dict[str, Location],
    ) -> None:
        """
        Initialize the class with the results of multiple cases of MultiCPTBearingResults.

        Parameters
        ----------
        results_per_case : Dict[Hashable, MultiCPTBearingResults]
            A dictionary with the results of multiple cases of MultiCPTBearingResults.
            The keys of the dictionary are the case names.
            All MultiCPTBearingResults objects must have the same pile tip levels and same test ids.
        cpt_locations : Dict[str, Location]
            The mapping between `test_id` (key) and the `location` (value) of the cpt.
            The keys must contain all the `test_id` (s) used in the MultiCPTBearingResults objects.

        Raises
        ------
        TypeError
            If `results_per_case` is not of the expected type.
            If `cpt_locations` is not of the expected type.
        ValueError
            If `results_per_case` is an empty dictionary.
            If not all MultiCPTBearingResults objects have the same pile tip levels and test ids.
            If not all the `test_id` (s) used in the MultiCPTBearingResults objects are in the keys of `cpt_locations`.
        """
        # Validate results cases
        _validate_results_per_case(results_per_case)

        # Initialize private variables
        self._cases = natsorted(list(results_per_case.keys()))
        self._multicpt_bearing_results = [results_per_case[c] for c in self.cases]
        self._test_ids = natsorted(
            results_per_case[list(results_per_case.keys())[0]].cpt_results.test_ids
        )
        self._pile_tip_levels_nap = sorted(
            list(
                results_per_case[list(results_per_case.keys())[0]]
                .cpt_results.to_pandas()
                .pile_tip_level_nap.unique()
            ),
            reverse=True,
        )
        self._set_cpt_locations(cpt_locations)

        # Create cpt_results_dataframe and cpt_group_results_dataframe
        self._set_cpt_results_dataframe(results_per_case)
        self._set_cpt_group_results_dataframe(results_per_case)

    def _set_cpt_results_dataframe(
        self, results_per_case: Dict[Hashable, MultiCPTBearingResults]
    ) -> None:
        """Private method to create and set the property `cpt_results_dataframe`."""
        records = []
        for case_name, case_results in results_per_case.items():
            for result_definition in CPTResultDefinitions:
                if (
                    result_definition.name
                    in case_results.cpt_results.to_pandas().columns
                ):
                    df = case_results.cpt_results.get_results_per_cpt(
                        column_name=result_definition.name
                    )
                    for idx_row, row in df.iterrows():
                        for test_id, result in row.items():
                            records.append(
                                dict(
                                    case_name=case_name,
                                    result_name=result_definition.name,
                                    test_id=test_id,
                                    x=self.cpt_locations[test_id].x,
                                    y=self.cpt_locations[test_id].y,
                                    pile_tip_level_nap=idx_row,
                                    result=result,
                                    result_unit=result_definition.value.unit,
                                )
                            )
        self._cpt_results_dataframe = pd.DataFrame.from_records(records)

    def _set_cpt_group_results_dataframe(
        self, result_cases: Dict[Hashable, MultiCPTBearingResults]
    ) -> None:
        """Private method to create and set the property `cpt_group_results_dataframe`."""
        records = []
        for case_name, case_results in result_cases.items():
            df = case_results.group_results_table.to_pandas()
            for result_definition in CPTGroupResultDefinitions:
                if result_definition.name in df.columns:
                    for _, row in df.iterrows():
                        records.append(
                            dict(
                                case_name=case_name,
                                result_name=result_definition.name,
                                pile_tip_level_nap=row["pile_tip_level_nap"],
                                result=row[result_definition.value.name],
                                result_unit=result_definition.value.unit,
                            )
                        )
        self._cpt_group_results_dataframe = pd.DataFrame.from_records(records)

    def _set_cpt_locations(self, value: Dict[str, Location]) -> None:
        """Private setter for `cpt_locations`."""
        # Check data types
        if not isinstance(value, dict):
            raise TypeError(
                f"Expected type 'List[CPTData]' for 'cpt_locations', but got {type(value)}"
            )

        if not all(isinstance(k, str) for k in value.keys()):
            raise TypeError(
                f"Expected type 'str' for keys of 'cpt_locations', but got {(type(k) for k in value.keys())}"
            )

        if not all(isinstance(v, Location) for v in value.values()):
            raise TypeError(
                f"Expected type 'Location' for values of 'cpt_locations', but got {(type(v) for v in value.values())}"
            )

        # Check that all the `test_id`s used in the MultiCPTBearingResults objects are in the keys.
        if not all(test_id in value.keys() for test_id in self.test_ids):
            missing_test_ids = [
                test_id for test_id in self.test_ids if test_id not in value.keys()
            ]
            raise ValueError(
                "Not all `test_id`s used in the MultiCPTBearingResults objects are in the keys of `cpt_locations`. "
                + f"The following `test_id`s are missing: {missing_test_ids}"
            )

        self._cpt_locations = value

    @property
    def cases(self) -> List[Hashable]:
        """The case names."""
        return self._cases

    @property
    def multicpt_bearing_results(self) -> List[MultiCPTBearingResults]:
        """The MultiCPTBearingResults objects."""
        return self._multicpt_bearing_results

    @property
    def test_ids(self) -> List[str]:
        """The test ids of all MultiCPTBearingResults objects."""
        return self._test_ids

    @property
    def pile_tip_levels_nap(self) -> List[float]:
        """The pile tip levels NAP of all MultiCPTBearingResults objects."""
        return self._pile_tip_levels_nap

    @property
    def cpt_locations(self) -> Dict[str, Location]:
        """The mapping between `test_id` (key) and the `location` (value) of the cpt."""
        return self._cpt_locations

    @property
    def cpt_results_dataframe(self) -> pd.DataFrame:
        """
        The dataframe with all the CPT results.
        Available columns: case_name, result_name, test_id, x, y, pile_tip_level_nap, result, result_unit.
        """
        return self._cpt_results_dataframe

    @property
    def cpt_group_results_dataframe(self) -> pd.DataFrame:
        """
        The dataframe with CPT group results.
        Available columns: case_name, result_name, pile_tip_level_nap, result, result_unit.
        """
        return self._cpt_group_results_dataframe


def _validate_results_per_case(
    results_per_case: Dict[Hashable, MultiCPTBearingResults]
) -> None:
    """
    Private method to validate the results_per_case dictionary.

    Parameters
    ----------
    results_per_case : Dict[Hashable, MultiCPTBearingResults]
        A dictionary with the results of multiple cases of MultiCPTBearingResults.
        The keys of the dictionary are the case names.
        All MultiCPTBearingResults objects must have the same pile tip levels and same test ids.

    Raises
    ------
    TypeError
        If `results_per_case` is not of the expected type.
    ValueError
        If `results_per_case` is an empty dictionary.
        If not all MultiCPTBearingResults objects have the same pile tip levels and test ids.
    """
    if not isinstance(results_per_case, dict):
        raise TypeError(
            f"Expected type 'Dict[Hashable, MultiCPTBearingResults]' for 'results_per_case', but got {type(results_per_case)}"
        )

    if len(results_per_case) == 0:
        raise ValueError("Empty dictionary 'results_per_case' is not allowed.")

    for val in results_per_case.values():
        if not isinstance(val, MultiCPTBearingResults):
            raise TypeError(
                f"Expected type 'MultiCPTBearingResults' for items in 'results_per_case', but got {type(val)}"
            )

    test_ids = results_per_case[list(results_per_case.keys())[0]].cpt_results.test_ids
    pile_tip_levels_nap = list(
        results_per_case[list(results_per_case.keys())[0]]
        .cpt_results.to_pandas()
        .pile_tip_level_nap.unique()
    )
    for results in results_per_case.values():
        if results.cpt_results.test_ids != test_ids:
            raise ValueError(
                "All MultiCPTBearingResults objects must have the same test ids."
            )

        if (
            list(results.cpt_results.to_pandas().pile_tip_level_nap.unique())
            != pile_tip_levels_nap
        ):
            raise ValueError(
                "All MultiCPTBearingResults objects must have the same pile tip levels."
            )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.typing import NDArray

from pypilecore.results.soil_properties import (
    CPTTable,
    LayerTable,
    SoilProperties,
    get_soil_layer_handles,
)

Number = Union[float, int]


@dataclass(frozen=True)
class CPTCompressionResultsTable:
    """Object containing the results of a single CPT."""

    pile_tip_level_nap: NDArray[np.float64]
    """The pile-tip level in [m] w.r.t. the reference."""
    F_nk_cal: NDArray[np.float64]
    """The calculated value of the negative shaft friction force [kN]."""
    F_nk_k: NDArray[np.float64]
    """The characteristic value of the negative shaft friction force [kN]."""
    F_nk_d: NDArray[np.float64]
    """The design value of the negative shaft friction force [kN]."""
    R_b_cal: NDArray[np.float64]
    """The calculated value of the bottom bearingcapacity [kN]."""
    R_b_k: NDArray[np.float64]
    """The characteristic value of the bottom bearingcapacity [kN]."""
    R_b_d: NDArray[np.float64]
    """The design value of the bottom bearingcapacity [kN]."""
    R_s_cal: NDArray[np.float64]
    """The calculated value of the shaft bearingcapacity [kN]."""
    R_s_k: NDArray[np.float64]
    """The characteristic value of the shaft bearingcapacity [kN]."""
    R_s_d: NDArray[np.float64]
    """The design value of the shaft bearingcapacity [kN]."""
    R_c_cal: NDArray[np.float64]
    """The calculated value of the total compressive bearingcapacity [kN]."""
    R_c_k: NDArray[np.float64]
    """The characteristic value of the total compressive bearingcapacity [kN]."""
    R_c_d: NDArray[np.float64]
    """The design value of the total compressive bearingcapacity [kN]."""
    R_c_d_net: NDArray[np.float64]
    """The net design value of the total bearingcapacity [kN] (netto =excluding design negative friction force.)."""
    F_c_k: NDArray[np.float64]
    """The compressive force on the pile-head [kN]."""
    F_c_k_tot: NDArray[np.float64]
    """The characteristic value of the total compressive pile load [kN](building-load + neg. friction force)."""
    negative_friction_range_nap_top: NDArray[np.float64]
    """The top boundary of the negative friction interval [m] w.r.t. NAP.
    Can be None when the friction force was provided directly."""
    negative_friction_range_nap_btm: NDArray[np.float64]
    """The bottom boundary of the negative friction interval [m] w.r.t. NAP.
    Can be None when the friction force was provided directly."""
    positive_friction_range_nap_top: NDArray[np.float64]
    """The top boundary of the positive friction interval [m] w.r.t. NAP."""
    positive_friction_range_nap_btm: NDArray[np.float64]
    """The bottom boundary of the positive friction interval [m] w.r.t. NAP."""
    q_b_max: NDArray[np.float64]
    """The maximum bottom bearing resistance [MPa]."""
    q_s_max_mean: NDArray[np.float64]
    """The maximum shaft bearing resistance [MPa]."""
    qc1: NDArray[np.float64]
    """The average friction resistance in Koppejan trajectory I, :math:`q_{c;I;gem}` [MPa] ."""
    qc2: NDArray[np.float64]
    """The average friction resistance in Koppejan trajectory II, :math:`q_{c;II;gem}` [MPa] ."""
    qc3: NDArray[np.float64]
    """The average friction resistance in Koppejan trajectory III, :math:`q_{c;III;gem}` [MPa] ."""
    s_b: NDArray[np.float64]
    """The settlement of the pile bottom [mm]."""
    s_el: NDArray[np.float64]
    """The elastic shortening of the pile due to elastic strain [mm]."""
    k_v_b: NDArray[np.float64]
    """The 1-dimensional stiffness modulus at pile bottom [kN/m]."""
    k_v_1: NDArray[np.float64]
    """The 1-dimensional stiffness modulus at pile head [MN/mm]."""

    def __post_init__(self) -> None:
        dict_lengths = {}
        for key, value in self.__dict__.items():
            if not np.all(np.isnan(value)):
                dict_lengths[key] = len(value)
        if len(set(dict_lengths.values())) > 1:
            raise ValueError(
                f"Inputs for LayerTable must have same lengths, but got lengths: {dict_lengths}"
            )

    @classmethod
    def from_sequences(
        cls,
        pile_tip_level_nap: Sequence[Number],
        F_nk_cal: Sequence[Number],
        F_nk_k: Sequence[Number],
        F_nk_d: Sequence[Number],
        R_b_cal: Sequence[Number],
        R_b_k: Sequence[Number],
        R_b_d: Sequence[Number],
        R_s_cal: Sequence[Number],
        R_s_k: Sequence[Number],
        R_s_d: Sequence[Number],
        R_c_cal: Sequence[Number],
        R_c_k: Sequence[Number],
        R_c_d: Sequence[Number],
        R_c_d_net: Sequence[Number],
        F_c_k: Sequence[Number],
        F_c_k_tot: Sequence[Number],
        negative_friction_range_nap_top: Sequence[Number],
        negative_friction_range_nap_btm: Sequence[Number],
        positive_friction_range_nap_top: Sequence[Number],
        positive_friction_range_nap_btm: Sequence[Number],
        q_b_max: Sequence[Number],
        q_s_max_mean: Sequence[Number],
        qc1: Sequence[Number],
        qc2: Sequence[Number],
        qc3: Sequence[Number],
        s_b: Sequence[Number],
        s_el: Sequence[Number],
        k_v_b: Sequence[Number],
        k_v_1: Sequence[Number],
    ) -> CPTCompressionResultsTable:
        return cls(
            pile_tip_level_nap=np.array(pile_tip_level_nap).astype(np.float64),
            F_nk_cal=np.array(F_nk_cal).astype(np.float64),
            F_nk_k=np.array(F_nk_k).astype(np.float64),
            F_nk_d=np.array(F_nk_d).astype(np.float64),
            R_b_cal=np.array(R_b_cal).astype(np.float64),
            R_b_k=np.array(R_b_k).astype(np.float64),
            R_b_d=np.array(R_b_d).astype(np.float64),
            R_s_cal=np.array(R_s_cal).astype(np.float64),
            R_s_k=np.array(R_s_k).astype(np.float64),
            R_s_d=np.array(R_s_d).astype(np.float64),
            R_c_cal=np.array(R_c_cal).astype(np.float64),
            R_c_k=np.array(R_c_k).astype(np.float64),
            R_c_d=np.array(R_c_d).astype(np.float64),
            R_c_d_net=np.array(R_c_d_net).astype(np.float64),
            F_c_k=np.array(F_c_k).astype(np.float64),
            F_c_k_tot=np.array(F_c_k_tot).astype(np.float64),
            negative_friction_range_nap_top=np.array(
                negative_friction_range_nap_top
            ).astype(np.float64),
            negative_friction_range_nap_btm=np.array(
                negative_friction_range_nap_btm
            ).astype(np.float64),
            positive_friction_range_nap_top=np.array(
                positive_friction_range_nap_top
            ).astype(np.float64),
            positive_friction_range_nap_btm=np.array(
                positive_friction_range_nap_btm
            ).astype(np.float64),
            q_b_max=np.array(q_b_max).astype(np.float64),
            q_s_max_mean=np.array(q_s_max_mean).astype(np.float64),
            qc1=np.array(qc1).astype(np.float64),
            qc2=np.array(qc2).astype(np.float64),
            qc3=np.array(qc3).astype(np.float64),
            s_b=np.array(s_b).astype(np.float64),
            s_el=np.array(s_el).astype(np.float64),
            k_v_b=np.array(k_v_b).astype(np.float64),
            k_v_1=np.array(k_v_1).astype(np.float64),
        )

    def to_pandas(self) -> pd.DataFrame:
        """Get the pandas.DataFrame representation"""
        return pd.DataFrame(self.__dict__).dropna(axis=0, how="all")


class SingleCPTCompressionBearingResults:
    """
    Object that contains the results of a PileCore single-cpt calculation.

    *Not meant to be instantiated by the user.*
    """

    def __init__(
        self,
        soil_properties: SoilProperties,
        pile_head_level_nap: float,
        results_table: CPTCompressionResultsTable,
    ) -> None:
        """
        Parameters
        ----------
        soil_properties
            The object with soil properties
        pile_head_level_nap
            The elevation of the pile-head, in [m] w.r.t. NAP.
        results_table
            The object with CPT results.
        """
        self._sp = soil_properties
        self._pile_head_level_nap = pile_head_level_nap
        self._results_table = results_table

    @classmethod
    def from_api_response(
        cls,
        cpt_results_dict: dict,
        ref_height: float,
        surface_level_ref: float,
        x: float | None = None,
        y: float | None = None,
    ) -> "SingleCPTCompressionBearingResults":
        results_table = cpt_results_dict["results_table"]
        return cls(
            soil_properties=SoilProperties(
                cpt_table=CPTTable.from_api_response(
                    cpt_results_dict.get("cpt_chart", {})
                ),
                layer_table=LayerTable.from_api_response(
                    cpt_results_dict["layer_table"]
                ),
                test_id=cpt_results_dict.get("test_id"),
                ref_height=ref_height,
                surface_level_ref=surface_level_ref,
                groundwater_level_ref=cpt_results_dict["groundwater_level_nap"],
                x=x,
                y=y,
            ),
            pile_head_level_nap=cpt_results_dict["annotations"]["pile_head_level_nap"],
            results_table=CPTCompressionResultsTable.from_sequences(
                pile_tip_level_nap=results_table["pile_tip_level_nap"],
                F_nk_cal=results_table["F_nk_cal"],
                F_nk_k=results_table["F_nk_k"],
                F_nk_d=results_table["F_nk_d"],
                R_b_cal=results_table["R_b_cal"],
                R_b_k=results_table["R_b_k"],
                R_b_d=results_table["R_b_d"],
                R_s_cal=results_table["R_s_cal"],
                R_s_k=results_table["R_s_k"],
                R_s_d=results_table["R_s_d"],
                R_c_cal=results_table["R_c_cal"],
                R_c_k=results_table["R_c_k"],
                R_c_d=results_table["R_c_d"],
                R_c_d_net=results_table["R_c_d_net"],
                F_c_k=results_table["F_c_k"],
                F_c_k_tot=results_table["F_c_k_tot"],
                negative_friction_range_nap_top=results_table[
                    "negative_friction_range_nap_top"
                ],
                negative_friction_range_nap_btm=results_table[
                    "negative_friction_range_nap_btm"
                ],
                positive_friction_range_nap_top=results_table[
                    "positive_friction_range_nap_top"
                ],
                positive_friction_range_nap_btm=results_table[
                    "positive_friction_range_nap_btm"
                ],
                q_b_max=results_table["q_b_max"],
                q_s_max_mean=results_table["q_s_max_mean"],
                qc1=results_table["qc1"],
                qc2=results_table["qc2"],
                qc3=results_table["qc3"],
                s_b=results_table["s_b"],
                s_el=results_table["s_el"],
                k_v_b=results_table["k_v_b"],
                k_v_1=results_table["k_v_1"],
            ),
        )

    @property
    def soil_properties(self) -> SoilProperties:
        """
        The SoilProperties object.
        """
        return self._sp

    @property
    def pile_head_level_nap(self) -> float:
        """
        The elevation of the pile-head in [m] w.r.t. NAP.
        """
        return self._pile_head_level_nap

    @property
    def table(self) -> CPTCompressionResultsTable:
        """The object with single-CPT results table traces."""
        return self._results_table

    def plot_bearing_capacities(
        self,
        axes: Optional[Axes] = None,
        figsize: Tuple[float, float] = (8, 10),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the bearing calculation results on an `Axes' object.

        Parameters
        ----------
        axes:
            Optional `Axes` object where the bearing capacities can be plotted on.
            If not provided, a new `plt.Figure` will be activated and the `Axes`
            object will be created and returned.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        add_legend:
            Add a legend to the second axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the bearing capacities were plotted on.
        """

        # Create axes objects if not provided
        if axes is not None:
            if not isinstance(axes, Axes):
                raise ValueError(
                    "'axes' argument to plot_bearing_capacities() must be a `pyplot.axes.Axes` object or None."
                )
        else:
            kwargs_subplot = {
                "figsize": figsize,
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(1, 1, **kwargs_subplot)

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        # add horizontal lines
        axes.axhline(
            y=self.soil_properties.groundwater_level_ref,
            color="tab:blue",
            linestyle="--",
            label="Groundwater level",
        )
        axes.axhline(
            y=self.soil_properties.surface_level_ref,
            color="tab:brown",
            linestyle="--",
            label="Surface level",
        )

        # add bearing result subplot
        axes.plot(
            np.array(self.table.F_nk_d),
            self.table.pile_tip_level_nap,
            color="tab:orange",
            label="Fnk;d",
        )
        axes.plot(
            np.array(self.table.R_s_cal),
            self.table.pile_tip_level_nap,
            color="lightgreen",
            label="Rs;cal;max",
        )
        axes.plot(
            np.array(self.table.R_b_cal),
            self.table.pile_tip_level_nap,
            color="darkgreen",
            label="Rb;cal;max",
        )
        axes.plot(
            np.array(self.table.R_c_d_net),
            self.table.pile_tip_level_nap,
            label=r"Rc;net;d",
            lw=3,
            color="tab:blue",
        )
        axes.set_xlabel("Force [kN]")

        # add legend
        if add_legend:
            axes.legend(
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )

        # set grid
        axes.grid()

        return axes

    def plot_bearing_overview(
        self,
        figsize: Tuple[float, float] = (10.0, 12.0),
        width_ratios: Tuple[float, float, float] = (1, 0.1, 2),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Figure:
        """
        Plot an overview of the bearing-capacities, including the .

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        width_ratios:
            Tuple of width-ratios of the subplots, as the `plt.GridSpec` argument.
        add_legend:
            Add a legend to the second axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        fig:
            The matplotlib Figure
        """

        kwargs_subplot = {
            "gridspec_kw": {"width_ratios": width_ratios},
            "sharey": "row",
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        fig, _ = plt.subplots(
            1,
            3,
            **kwargs_subplot,
        )

        ax_qc, ax_layers, ax_bearing = fig.axes
        ax_rf = ax_qc.twiny()
        assert isinstance(ax_rf, Axes)

        # Plot bearing capacities
        self.soil_properties.cpt_table.plot_qc(ax_qc, add_legend=False)
        self.soil_properties.cpt_table.plot_friction_ratio(ax_rf, add_legend=False)
        self.soil_properties.plot_layers(ax_layers, add_legend=False)
        self.plot_bearing_capacities(axes=ax_bearing, add_legend=False)

        if add_legend:
            ax_qc_legend_handles_list = ax_qc.get_legend_handles_labels()[0]
            ax_rf_legend_handles_list = ax_rf.get_legend_handles_labels()[0]
            ax_layers_legend_handles_list = get_soil_layer_handles()

            # Omit last 2 duplicate "bearing" handles
            # (groundwater_level and surface_level):
            ax_bearing_legend_handles_list = ax_bearing.get_legend_handles_labels()[0][
                2:
            ]

            handles_list = [
                *ax_qc_legend_handles_list,
                *ax_rf_legend_handles_list,
                *ax_layers_legend_handles_list,
                *ax_bearing_legend_handles_list,
            ]

            ax_bearing.legend(
                handles=handles_list,
                loc="upper left",
                bbox_to_anchor=(1, 1),
                title="name: " + self.soil_properties.test_id
                if self.soil_properties.test_id is not None
                else "name: unknown",
            )

        return fig



from __future__ import annotations

from functools import lru_cache
from typing import Any, List, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Patch
from numpy.typing import NDArray

from pypilecore.utils import depth_to_nap, nap_to_depth

Number = Union[float, int]

SOIL_COLOR_DIC_intern = {
    "G": "#708090",
    "Z": "#DBAD4B",
    "L": "#0078C1",
    "K": "#578E57",
    "V": "#a76b29",
}

ENG_MAIN_COMPONENT_NAME_DIC = {
    "G": "Gravel",
    "Z": "Sand",
    "L": "Loam",
    "V": "Peat",
    "K": "Clay",
}

SOIL_COLOR_DIC = {
    value: SOIL_COLOR_DIC_intern[key]
    for key, value in ENG_MAIN_COMPONENT_NAME_DIC.items()
}


def get_soil_layer_handles() -> List[Patch]:
    return [Patch(color=clr, label=key) for (key, clr) in SOIL_COLOR_DIC.items()]


class LayerTable:
    """
    Object that contains the Soil-layer data-traces.
    """

    def __init__(
        self,
        index: Sequence[int],
        thickness: Sequence[float],
        depth_btm: Sequence[float],
        C_s: Sequence[float] | None,
        C_p: Sequence[float] | None,
        gamma: Sequence[float],
        gamma_sat: Sequence[float],
        phi: Sequence[float],
        soil_code: Sequence[str],
    ):
        self.index = np.array(index).astype(np.float64)
        """Layer index"""
        self.thickness = np.array(thickness).astype(np.float64)
        """The layer thickness [m]"""
        self.depth_btm = np.array(depth_btm).astype(np.float64)
        """The depth of the layer bottom (below service level) [m]."""
        self.C_s = np.array(C_s).astype(np.float64)
        """Koppejan parameters for secondary compression."""
        self.C_p = np.array(C_p).astype(np.float64)
        """Koppejan parameters for primary compression."""
        self.gamma = np.array(gamma).astype(np.float64)
        """The dry unit weights [MPa]."""
        self.gamma_sat = np.array(gamma_sat).astype(np.float64)
        """The saturated unit weights [MPa]."""
        self.phi = np.array(phi).astype(np.float64)
        """Internal friction angle. [rad]"""
        self.soil_code = np.array(soil_code).astype(np.str_)
        """
        The code used to describe the soil layers of the boreholes. Main components are
        specified with capital letters and are the following:

            - G: gravel (Grind)
            - Z: sand (Zand)
            - L: loam (Leem)
            - K: clay (Klei)
            - V: peat (Veen)
        """

        dict_lengths = {}
        for key, value in self.__dict__.items():
            if not np.all(pd.isnull(value)):
                dict_lengths[key] = len(value)
        if len(set(dict_lengths.values())) > 1:
            raise ValueError(
                f"Inputs for LayerTable must have same lengths, but got lengths: {dict_lengths}"
            )

        self.__dict__.update({"depth_top": self.depth_top})

    @classmethod
    def from_api_response(cls, layer_table_dict: dict) -> "LayerTable":
        """
        Instantiates the LayerTable object from the "layer_table" object, which is returned in
        the response of a "compression/multiple-cpts/results" endpoint call.
        """
        return cls(
            index=layer_table_dict["index"],
            thickness=layer_table_dict["thickness"],
            depth_btm=layer_table_dict["depth_btm"],
            C_s=layer_table_dict.get("C_s"),
            C_p=layer_table_dict.get("C_p"),
            gamma=layer_table_dict["gamma"],
            gamma_sat=layer_table_dict["gamma_sat"],
            phi=layer_table_dict["phi"],
            soil_code=layer_table_dict["soil_code"],
        )

    @property
    def depth_top(self) -> NDArray[np.float64]:
        return self.depth_btm - self.thickness

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """The pandas.DataFrame representation"""
        return pd.DataFrame(self.__dict__).dropna(axis=0, how="all")


class CPTTable:
    """
    Object that contains the CPT-related data-traces of a bearing calculation. These
    can be either raw input data, corrected data or intermediate results.
    """

    def __init__(
        self,
        depth_nap: Sequence[float] | None,
        qc: Sequence[float] | None,
        qc_original: Sequence[float] | None,
        qc_chamfered: Sequence[float] | None,
        qc1: Sequence[float] | None,
        qc2: Sequence[float] | None,
        fs: Sequence[float] | None,
        qs_d: Sequence[float] | None,
        f1: Sequence[float] | None,
        f2_d_mean: Sequence[float] | None,
        f3: Sequence[float] | None,
        R_t_d: Sequence[float] | None,
        phi_plug: Sequence[float] | None,
        R_t_d_plug: Sequence[float] | None,
        alpha_t: Sequence[float] | None,
        alpha_t_1: Sequence[float] | None,
        alpha_t_2: Sequence[float] | None,
    ):
        if depth_nap is None:
            self.depth_nap = np.array([depth_nap]).astype(np.float64)
        else:
            self.depth_nap = np.array(depth_nap).astype(np.float64).round(decimals=2)

        """The depth [m] w.r.t. NAP"""
        self.qc = np.array(qc).astype(np.float64)
        """The cone resistance signal from the CPT [MPa], possibly corrected for excavation
        or OCR."""
        self.qc_original = np.array(qc_original).astype(np.float64)
        """The original cone resistance signal from the CPT [MPa]."""
        self.qc_chamfered = np.array(qc_chamfered).astype(np.float64)
        """The chamfered-qc signal, used for the positive friction range [MPa]."""
        self.qc1 = np.array(qc1).astype(np.float64)
        """The Koppejan-qc1 trajectory [MPa]."""
        self.qc2 = np.array(qc2).astype(np.float64)
        """The Koppejan-qc2 trajectory [MPa]."""
        self.fs = np.array(fs).astype(np.float64)
        """The original fs signal from the CPT [MPa]."""
        self.qs_d = np.array(qs_d).astype(np.float64)
        """The computational value of shaft friction [kPa]."""
        self.f1 = np.array(f1).astype(np.float64)
        """factor for the effect of compaction in the pile group (7.6.3.3 (e)
            NEN 9997-1+C2:2017) [-]."""
        self.f2_d_mean = np.array(f2_d_mean).astype(np.float64)
        """factor for the decrease in grain stress in sand layers from which
        the pile derives its tensile resistance (7.6.3.3 (f) NEN
        9997-1+C2:2017) [-]."""
        self.f3 = np.array(f3).astype(np.float64)
        """factor for the decrease in bearing capacity from the length—effect,
        according to figure 6.1a en 6.1b (CROW-CUR Rapport 236   Richtlijn
        Ankerpalen [2023]) [-]."""
        self.R_t_d = np.array(R_t_d).astype(np.float64)
        """calculation value of the tensile resistance of a pile or pile group
        (7.6.3.3 (a) NEN 9997-1+C2:2017 or CROW-CUR Rapport 236 Richtlijn Ankerpalen [2023]) [kN]."""
        self.phi_plug = np.array(phi_plug).astype(np.float64)
        """Half top angle of the ground cone (7.6.3.3 Table 7.e NEN 9997-1+C2:2017) [degrees]."""
        self.R_t_d_plug = np.array(R_t_d_plug).astype(np.float64)
        """root ball weight, excluding the weight of the pile (7.6.3.3 (h) NEN
        9997-1+C2:2017) [kN]."""
        self.alpha_t = np.array(alpha_t).astype(np.float64)
        """Alpha t factor used in tension resistance calculation [-]. based on
        table 7.d NEN 9997-1+C2:2017."""
        self.alpha_t_1 = np.array(alpha_t_1).astype(np.float64)
        """Alpha t factor used in tension resistance calculation [-]. tabel 6.2
        CROW-CUR Rapport 236 Richtlijn Ankerpalen [2023]."""
        self.alpha_t_2 = np.array(alpha_t_2).astype(np.float64)
        """Alpha t factor used in tension resistance calculation [-]. tabel 6.2
        CROW-CUR Rapport 236 Richtlijn Ankerpalen [2023]."""

        dict_lengths = {}
        for key, value in self.__dict__.items():
            if not np.all(np.isnan(value)):
                dict_lengths[key] = len(value)
        if len(set(dict_lengths.values())) > 1:
            raise ValueError(
                f"Inputs for CPTTable must have same lengths, but got lengths: {dict_lengths}"
            )

        self.__dict__.update({"friction_ratio": self.friction_ratio})

    @classmethod
    def from_api_response(cls, cpt_chart_dict: dict) -> "CPTTable":
        """
        Instantiates the CPTTable object from the "cpt_chart" object, which is returned in
        the response of a "compression/multiple-cpts/results" endpoint call.
        """
        return cls(
            depth_nap=cpt_chart_dict.get("depth_nap"),
            qc=cpt_chart_dict.get("qc"),
            qc_original=cpt_chart_dict.get("qc_original"),
            qc_chamfered=cpt_chart_dict.get("qc_chamfered"),
            qc1=cpt_chart_dict.get("qc1"),
            qc2=cpt_chart_dict.get("qc2"),
            fs=cpt_chart_dict.get("fs"),
            qs_d=cpt_chart_dict.get("qs_d"),
            f1=cpt_chart_dict.get("f1"),
            f2_d_mean=cpt_chart_dict.get("f2_d_mean"),
            f3=cpt_chart_dict.get("f3"),
            R_t_d=cpt_chart_dict.get("R_t_d"),
            phi_plug=cpt_chart_dict.get("phi_plug"),
            R_t_d_plug=cpt_chart_dict.get("R_t_d_plug"),
            alpha_t=cpt_chart_dict.get("alpha_t"),
            alpha_t_1=cpt_chart_dict.get("alpha_t_1"),
            alpha_t_2=cpt_chart_dict.get("alpha_t_2"),
        )

    @property
    def friction_ratio(self) -> NDArray[np.float64]:
        return np.array(self.fs / self.qc * 100)

    @property
    def qc_has_been_chamfered(self) -> bool:
        """Returns False if `qc_chamfered` contains the same data as `qc`."""
        return not np.allclose(self.qc, self.qc_chamfered)

    @property
    def qc_has_been_reduced(self) -> bool:
        """Returns False if `qc` contains the same data as `qc_original`."""
        return not np.allclose(self.qc_original, self.qc)

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """The pandas.DataFrame representation"""
        return pd.DataFrame(self.__dict__).dropna(axis=0, how="all")

    def plot_qc(
        self,
        axes: Axes | None = None,
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plots the qc data on the provided `Axes`.

        Parameters
        ----------
        axes:
            Optional Axes to plot on.
        add_legend:
            Add a legend (default = True).
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The matplotlib Axes object
        """

        if axes is not None:
            if not isinstance(axes, Axes):
                raise TypeError(
                    f"`axes` input for CPTTable.plot_qc() should be a `matplotlib.axes.Axes` object or None, but got: {type(axes)}."
                )
        else:
            kwargs_subplot = {
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(
                1,
                1,
                **kwargs_subplot,
            )

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        if np.all(np.isnan(self.qc)):
            qc = np.ones_like(self.depth_nap) * np.nan
        else:
            qc = self.qc

        # Plot Base qc subplot
        if self.qc_has_been_chamfered is True:
            axes.plot(
                self.qc_chamfered,
                self.depth_nap,
                label="$q_{c;a}$",
                color="orange",
                linestyle=":",
            )

        if self.qc_has_been_reduced is True:
            axes.plot(
                self.qc_original,
                self.depth_nap,
                label="$q_{c;original}$",
                linestyle="-",
                color="darkblue",
            )
            axes.plot(
                qc,
                self.depth_nap,
                label="$q_{c;reduced}$",
                linestyle="-",
                color="orange",
            )

        else:
            axes.plot(
                qc,
                self.depth_nap,
                label="$q_c$",
                linestyle="-",
                color="darkblue",
            )

        axes.set_xlim((0, 40))
        axes.set_ylabel("Depth [m NAP]")
        axes.set_xlabel("$q_c$ [MPa]")
        axes.xaxis.label.set_color("darkblue")

        # add grid
        axes.grid()

        if add_legend:
            axes.legend(
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )

        return axes

    def plot_friction_ratio(
        self,
        axes: Axes | None = None,
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plots the friction-ratio data on the provided `Axes`.

        Parameters
        ----------
        axes:
            Optional Axes to plot on.
        add_legend:
            Add a legend (default = True).
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The matplotlib Axes object
        """

        if axes is not None:
            if not isinstance(axes, Axes):
                raise TypeError(
                    f"`axes` input for CPTTable.plot_rf() should be a `matplotlib.axes.Axes` object or None, but got: {type(axes)}."
                )
        else:
            kwargs_subplot = {
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(
                1,
                1,
                **kwargs_subplot,
            )

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        if np.all(np.isnan(self.friction_ratio)):
            friction_ratio = np.ones_like(self.depth_nap) * np.nan
        else:
            friction_ratio = self.friction_ratio

        # add friction number subplot
        axes.plot(
            friction_ratio,
            self.depth_nap,
            label="Rf",
            color="darkgray",
        )
        axes.spines["top"].set_position(("outward", 0))
        axes.set_xlabel("Friction ratio [%]")
        axes.xaxis.label.set_color("lightgray")

        axes.set_xlim(0, 20)

        if add_legend:
            axes.legend(
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )

        return axes


class SoilProperties:
    """
    A class for soil properties.
    """

    def __init__(
        self,
        cpt_table: CPTTable,
        layer_table: LayerTable,
        ref_height: float,
        surface_level_ref: float,
        groundwater_level_ref: float,
        test_id: str | None = None,
        x: float | None = None,
        y: float | None = None,
    ):
        """
        Parameters
        ----------
        cpt_table:
            The CPTTable object
        layer_table:
            The LayerTable object.
        ref_height:
            The vertical reference [m].
        surface_level_ref:
            The elevation of the surface w.r.t. the vertical reference [m]. This could
            be the level post-excavation.
        groundwater_level_ref:
            The elevation of the groundwater w.r.t. the vertical reference [m].
        test_id:
            Identifier of the CPT
        x:
            x coordinate of CPT
        y:
            y coordinate of CPT
        """
        self._cpt_table = cpt_table
        self._layer_table = layer_table
        self._ref_height = ref_height
        self._test_id = test_id
        self._groundwater_level_ref = groundwater_level_ref
        self._surface_level_ref = surface_level_ref
        self._x = x
        self._y = y

    @property
    def cpt_table(self) -> CPTTable:
        """The CPTTable object"""
        return self._cpt_table

    @property
    def layer_table(self) -> LayerTable:
        """The LayerTable object"""
        return self._layer_table

    @property
    def x(self) -> float | None:
        """x-coordinate of the CPT"""
        return self._x

    @property
    def y(self) -> float | None:
        """y-coordinate of the CPT"""
        return self._y

    @property
    def test_id(self) -> str | None:
        """Identifier of the CPT"""
        return self._test_id

    @property
    def ref_height(self) -> float:
        """The vertical reference [m]."""
        return self._ref_height

    @property
    def groundwater_level_ref(self) -> float:
        """The elevation of the groundwater w.r.t. the vertical reference [m]."""
        return self._groundwater_level_ref

    @property
    def surface_level_ref(self) -> float:
        """The elevation of the surface w.r.t. the vertical reference [m]. This could
        be the level post-excavation."""
        return self._surface_level_ref

    def plot_layers(
        self,
        axes: Axes | None = None,
        hide_excavated: bool = False,
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plots the soil layers on the provided `Axes`.

        Parameters
        ----------
        axes:
            Optional `Axes` object where the soil-layer data can be plotted on. If
            not provided, a new `plt.Figure` will be activated and the `Axes` object
            will be created and returned.
        hide_excavated:
            Hide the layers under the excavation level.
        add_legend:
            Add a legend to the axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the soil layers were plotted on
        """
        if axes is not None:
            if not isinstance(axes, Axes):
                raise TypeError(
                    f"`axes` input for CPTTable.plot_qc() should be a `matplotlib.axes.Axes` object or None, but got: {type(axes)}."
                )
        else:
            kwargs_subplot = {
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(
                1,
                1,
                **kwargs_subplot,
            )

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        # add soil layers subplot
        for depth_btm, delta_z, main_component in zip(
            self.layer_table.depth_btm,
            self.layer_table.thickness,
            self.layer_table.soil_code,
        ):
            if hide_excavated:
                if depth_btm < nap_to_depth(self.surface_level_ref, self.ref_height):
                    continue
                if depth_btm - delta_z < nap_to_depth(
                    self.surface_level_ref, self.ref_height
                ):
                    delta_z = depth_btm - nap_to_depth(
                        self.surface_level_ref, self.ref_height
                    )

            if main_component[0] not in list(SOIL_COLOR_DIC_intern.keys()):
                raise ValueError(
                    "Cannot plot Soil Properties, update SOIL_COLOR_DIC"
                    "to match soil_code of the layer table"
                )
            axes.fill_between(
                [0, 1],
                y1=depth_to_nap(depth_btm, self.ref_height) + delta_z,
                y2=depth_to_nap(depth_btm, self.ref_height),
                color=SOIL_COLOR_DIC_intern[main_component[0]],
            )
        axes.get_xaxis().set_visible(False)

        if add_legend:
            axes.legend(
                handles=get_soil_layer_handles(),
                loc="upper left",
                bbox_to_anchor=(1, 1),
                title="name: " + self.test_id
                if self.test_id is not None
                else "name: unknown",
            )

        return axes

    def plot(
        self,
        figsize: Tuple[float, float] = (10.0, 12.0),
        width_ratios: Tuple[float, float] = (1.0, 0.1),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Figure:
        """
        Plots the CPT and soil table data.

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        width_ratios:
            Tuple of width-ratios of the subplots, as the `plt.GridSpec` argument.
        add_legend:
            Add a legend to the second axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        fig:
            The matplotlib Figure
        """

        kwargs_subplot = {
            "gridspec_kw": {"width_ratios": width_ratios},
            "sharey": "row",
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        fig, _ = plt.subplots(
            1,
            2,
            **kwargs_subplot,
        )

        ax_qc, ax_layers = fig.axes

        ax_rf = ax_qc.twiny()
        assert isinstance(ax_rf, Axes)

        # Plot horizontal lines
        ax_qc.axhline(
            y=self.groundwater_level_ref,
            color="tab:blue",
            linestyle="--",
            label="Groundwater level",
        )

        ax_qc.axhline(
            y=self.surface_level_ref,
            color="tab:brown",
            linestyle="--",
            label="Surface level",
        )

        self.cpt_table.plot_qc(ax_qc, add_legend=False)
        self.cpt_table.plot_friction_ratio(ax_rf, add_legend=False)
        self.plot_layers(axes=ax_layers, add_legend=False)

        if add_legend:
            ax_qc_legend_handles_list = ax_qc.get_legend_handles_labels()[0]

            ax_rf_legend_handles_list = ax_rf.get_legend_handles_labels()[0]

            handles_list = [
                *ax_qc_legend_handles_list,
                *ax_rf_legend_handles_list,
            ]

            ax_qc.legend(
                handles=handles_list,
                loc="upper left",
                bbox_to_anchor=(1, 1),
                title="name: " + self.test_id
                if self.test_id is not None
                else "name: unknown",
            )

        # Add a legend if required
        if add_legend:
            ax_qc_legend_handles_list = ax_qc.get_legend_handles_labels()[0]

            ax_rf_legend_handles_list = ax_rf.get_legend_handles_labels()[0]

            soil_layer_legend_handles_list = get_soil_layer_handles()

            handles_list = [
                *ax_qc_legend_handles_list,
                *ax_rf_legend_handles_list,
                *soil_layer_legend_handles_list,
            ]

            ax_layers.legend(
                handles=handles_list,
                loc="upper left",
                bbox_to_anchor=(1, 1),
                title="name: " + self.test_id
                if self.test_id is not None
                else "name: unknown",
            )

        return fig



from __future__ import annotations

import math
from typing import Any, Dict, Literal, Tuple

import matplotlib.patches as patches
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy.typing import NDArray

from pypilecore.common.piles.geometry.components.common import (
    PrimaryPileComponentDimension,
    _BasePileGeometryComponent,
    get_area_vs_depth,
    get_circum_vs_depth,
    get_component_bounds_nap,
    instantiate_axes,
)


class RoundPileGeometryComponent(_BasePileGeometryComponent):
    """The RoundPileGeometryComponent class represents a round pile-geometry component."""

    def __init__(
        self,
        diameter: float,
        primary_dimension: PrimaryPileComponentDimension,
        inner_component: _BasePileGeometryComponent | None = None,
        material: str | None = None,
    ):
        """
        Represents a round pile-geometry component.

        Parameters:
        -----------
        diameter : float
            The outer-diameter [m] of the pile-geometry component.
        primary_dimension : PrimaryPileComponentDimension
            The primary dimension [m] of the pile-geometry component, which is measured along the primary axis of the pile.
        inner_component : RoundPileGeometryComponent | RectPileGeometryComponent | None, optional
            The component on the inside of the pile-geometry component, by default None.
        material : str, optional
            The material name of the pile-geometry component, by default None.
        """
        self._diameter = diameter
        self._primary_dimension = primary_dimension
        self._inner_component = inner_component
        self._material = material

    @classmethod
    def from_api_response(
        cls,
        component: dict,
        inner_component: _BasePileGeometryComponent | None = None,
    ) -> RoundPileGeometryComponent:
        """
        Instantiates a RoundPileGeometryComponent from a component object in the API
        response payload.

        Parameters:
        -----------
        component: dict
            A dictionary that represents the component object retrieved from the API response payload.
            The dictionary should have the following schema:
            {
                "diameter": float,  # The diameter of the round pile component.
                "primary_dimension": {
                    "length": float,  # The length of the primary dimension of the round pile component.
                },
                "material": str  # The material of the round pile component.
            }
        inner_component: RoundPileGeometryComponent | RectPileGeometryComponent | None, optional
            The component on the inside of the pile-geometry component, by default None.

        Returns:
        --------
        RoundPileGeometryComponent
            A round pile-geometry component.

        Example:
        --------
        >>> component = {
        ...     "diameter": 10,
        ...     "primary_dimension": {
        ...         "length": 20,
        ...         "width": 30
        ...     },
        ...     "material": "concrete"
        ... }
        >>> inner_component = RectPileGeometryComponent(...)
        >>> round_component = RoundPileGeometryComponent.from_api_response(component, inner_component)
        """
        return cls(
            diameter=component["diameter"],
            primary_dimension=PrimaryPileComponentDimension.from_api_response(
                component["primary_dimension"]
            ),
            inner_component=inner_component,
            material=component["material"],
        )

    @property
    def inner_component(
        self,
    ) -> _BasePileGeometryComponent | None:
        """The component on the inside of the pile-geometry component"""
        return self._inner_component

    @property
    def outer_shape(self) -> Literal["round"]:
        """The outer shape of the pile-geometry component"""
        return "round"

    @property
    def material(self) -> str | None:
        """The material name of the pile-geometry component"""
        return self._material

    @property
    def primary_dimension(self) -> PrimaryPileComponentDimension:
        """
        The primary dimension [m] of the pile-geometry component, which is measured along the primary axis of the pile.
        """
        return self._primary_dimension

    @property
    def cross_section_bounds(self) -> Tuple[float, float, float, float]:
        """Alias of the diameter [m] of the pile-geometry component"""
        return (
            -self.diameter / 2,
            self.diameter / 2,
            -self.diameter / 2,
            self.diameter / 2,
        )

    @property
    def diameter(self) -> float:
        """The outer-diameter [m] of the pile-geometry component"""
        return self._diameter

    @property
    def radius(self) -> float:
        """The outer-radius [m] of the pile-geometry component"""
        return self.diameter / 2

    @property
    def circumference(self) -> float:
        """The outer-circumference [m] of the pile-geometry component"""
        return self.diameter * math.pi

    @property
    def equiv_tip_diameter(self) -> float:
        """
        Equivalent diameter [m] of the component at tip-level.
        """
        return self.diameter

    @property
    def area_full(self) -> float:
        """The full outer-area [m²] of the pile-geometry component, including any potential inner-components"""
        return (self.diameter / 2) ** 2 * math.pi

    def serialize_payload(
        self,
    ) -> Dict[str, str | float | Dict[str, float | None] | None]:
        """
        Serialize the round pile-geometry component to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the outer shape, diameter, material, and primary dimension (if set).
        """
        return {
            "outer_shape": self.outer_shape,
            "primary_dimension": self.primary_dimension.serialize_payload(),
            "diameter": self.diameter,
            "material": self.material,
        }

    def get_component_bounds_nap(
        self,
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> Tuple[float, float]:
        """
        Returns component head and tip level in NAP.

        Parameters
        ----------
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        tuple
            Tuple with component head and tip level in [m] w.r.t. NAP.
        """
        return get_component_bounds_nap(
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
            component_primary_length=self.primary_dimension.length,
        )

    def get_circum_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns component circumferences at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with component circumferences at the depths in the depth parameter.
        """
        return get_circum_vs_depth(
            depth_nap=depth_nap,
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
            length=self.primary_dimension.length,
            circumference=self.circumference,
        )

    def get_inner_area_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns inner component areas at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with inner component areas at the depths in the depth parameter.
        """
        if self.inner_component is None:
            return np.zeros_like(depth_nap)

        return self.inner_component.get_area_vs_depth(
            depth_nap=depth_nap,
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
        )

    def get_area_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns component areas at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with component areas at the depths in the depth parameter.
        """

        (
            component_head_level_nap,
            component_tip_level_nap,
        ) = self.get_component_bounds_nap(pile_tip_level_nap, pile_head_level_nap)

        inner_area = self.get_inner_area_vs_depth(
            depth_nap=depth_nap,
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
        )

        return get_area_vs_depth(
            depth_nap=depth_nap,
            area_full=self.area_full,
            component_head_level_nap=component_head_level_nap,
            component_tip_level_nap=component_tip_level_nap,
            inner_area=inner_area,
        )

    def plot_cross_section_exterior(
        self,
        figsize: Tuple[float, float] = (6.0, 6.0),
        facecolor: Tuple[float, float, float] | str | None = None,
        axes: Axes | None = None,
        axis_arg: bool | str | Tuple[float, float, float, float] | None = "auto",
        show: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the cross-section of the component at a specified depth.

        Parameters
        ----------
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        facecolor : tuple or str, optional
            The face color of the pile cross-section, by default None.
        axes : Axes
            The axes object to plot the cross-section on.
        axis_arg : bool or str or tuple, optional
            The axis argument to pass to the `axes.axis()` function, by default "auto".
        show : bool, optional
            Whether to display the plot, by default True.
        **kwargs
            Additional keyword arguments to pass to the `plt.subplots()` function.
        """
        axes = instantiate_axes(
            figsize=figsize,
            axes=axes,
            **kwargs,
        )
        axes.add_patch(
            patches.Circle((0, 0), self.radius, facecolor=facecolor, edgecolor="black")
        )
        if axis_arg:
            axes.axis(axis_arg)
        if show:
            plt.show()
        return axes

    def plot_side_view(
        self,
        bottom_boundary_nap: float | Literal["pile_tip"] = "pile_tip",
        top_boundary_nap: float | Literal["pile_head"] = "pile_head",
        pile_tip_level_nap: float | int = -10,
        pile_head_level_nap: float | int = 0,
        figsize: Tuple[float, float] = (6.0, 6.0),
        facecolor: Tuple[float, float, float] | str | None = None,
        axes: Axes | None = None,
        axis_arg: bool | str | Tuple[float, float, float, float] | None = "scaled",
        show: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the side view of the component at a specified depth.

        Parameters
        ----------
        bottom_boundary_nap : float or str, optional
            The bottom boundary level of the plot, in m w.r.t. NAP. Default = "pile_tip".
        top_boundary_nap : float or str, optional
            The top boundary level of the plot, in m w.r.t. NAP. Default = "pile_head".
        pile_tip_level_nap : float, optional
            The pile tip level in m w.r.t. NAP. Default = -10.
        pile_head_level_nap : float, optional
            The pile head level in m w.r.t. NAP. Default = 0.
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        facecolor : tuple or str, optional
            The face color of the pile cross-section, by default None.
        axes : Axes
            The axes object to plot the cross-section on.
        axis_arg : bool or str or tuple, optional
            The axis argument to pass to the `axes.axis()` function, by default "auto".
        show : bool, optional
            Whether to display the plot, by default True.

        Returns
        -------
        Axes
            The axes object to plot the cross-section on.
        """
        axes = instantiate_axes(
            figsize=figsize,
            axes=axes,
            **kwargs,
        )

        if top_boundary_nap == "pile_head":
            top_boundary_nap = pile_head_level_nap

        if bottom_boundary_nap == "pile_tip":
            bottom_boundary_nap = pile_tip_level_nap

        (
            component_head_level_nap,
            component_tip_level_nap,
        ) = self.get_component_bounds_nap(pile_tip_level_nap, pile_head_level_nap)

        if (
            top_boundary_nap > component_tip_level_nap
            and bottom_boundary_nap < component_head_level_nap
        ):
            z_offset = component_head_level_nap
            height = (
                max(component_tip_level_nap, bottom_boundary_nap)
                - component_head_level_nap
            )

            axes.add_patch(
                patches.Rectangle(
                    (self.cross_section_bounds[0], z_offset),
                    self.diameter,
                    height,
                    facecolor=facecolor,
                )
            )

        if axis_arg:
            axes.axis(axis_arg)
        if show:
            plt.show()
        return axes



from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple

import matplotlib.patches as patches
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.collections import PatchCollection
from matplotlib.figure import Figure
from numpy.typing import NDArray
from scipy.spatial import Delaunay, Voronoi, voronoi_plot_2d

from pypilecore.results.soil_properties import SoilProperties, get_soil_layer_handles


class MaxBearingTable:
    """
    Object that contains the results belonging to the maximum net design bearing capacity (R_c_d_net) for a single CPT.

    *Not meant to be instantiated by the user.*
    """

    def __init__(
        self,
        pile_tip_level_nap: Sequence[float],
        R_c_d_net: Sequence[float],
        F_nk_d: Sequence[float],
        origin: Sequence[str],
    ):
        """
        Object that contains the results belonging to the maximum net design bearing capacity (R_c_d_net) for a single CPT.

        Parameters
        ----------
        pile_tip_level_nap
            The elevation of the pile-tip, in [m] w.r.t. NAP.
        R_c_d_net
            The maximum net design bearing capacity, in [kN].
        F_nk_d
            The net design bearing capacity, in [kN].
        origin
            The origin of the CPT data.
        """
        self._pile_tip_level_nap = pile_tip_level_nap
        self._R_c_d_net = R_c_d_net
        self._F_nk_d = F_nk_d
        self._origin = origin

    @property
    def pile_tip_level_nap(self) -> NDArray[np.float64]:
        """The elevation of the pile-tip, in [m] w.r.t. NAP."""
        return np.array(self._pile_tip_level_nap).astype(np.float64)

    @property
    def R_c_d_net(self) -> NDArray[np.float64]:
        """The maximum net design bearing capacity, in [kN]."""
        return np.array(self._R_c_d_net).astype(np.float64)

    @property
    def F_nk_d(self) -> NDArray[np.float64]:
        """The net design bearing capacity, in [kN]."""
        return np.array(self._F_nk_d).astype(np.float64)

    @property
    def origin(self) -> NDArray[np.str_]:
        """The origin of the CPT data."""
        return np.array(self._origin).astype(np.str_)

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """Get the pandas.DataFrame representation"""
        return pd.DataFrame(
            dict(
                pile_tip_level_nap=self.pile_tip_level_nap,
                R_c_d_net=self.R_c_d_net,
                F_nk_d=self.F_nk_d,
                origin=self.origin,
            )
        )


@dataclass(frozen=True)
class MaxBearingResult:
    """
    Object that contains the results of a PileCore single-cpt calculation.

    *Not meant to be instantiated by the user.*

    Attributes
    ----------
    soil_properties
        The object with soil properties
    pile_head_level_nap
        The elevation of the pile-head, in [m] w.r.t. NAP.
    table
        The object with CPT results.
    """

    soil_properties: SoilProperties
    pile_head_level_nap: float
    table: MaxBearingTable

    def to_pandas(self) -> pd.DataFrame:
        """Get the pandas.DataFrame representation"""
        return self.table.to_pandas()

    def plot_bearing_capacities(
        self,
        axes: Optional[Axes] = None,
        figsize: Tuple[float, float] = (8, 10),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the bearing calculation results on an `Axes' object.

        Parameters
        ----------
        axes:
            Optional `Axes` object where the bearing capacities can be plotted on.
            If not provided, a new `plt.Figure` will be activated and the `Axes`
            object will be created and returned.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        add_legend:
            Add a legend to the second axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the bearing capacities were plotted on.
        """

        # Create axes objects if not provided
        if axes is not None:
            if not isinstance(axes, Axes):
                raise ValueError(
                    "'axes' argument to plot_bearing_capacities() must be a `pyplot.axes.Axes` object or None."
                )
        else:
            kwargs_subplot = {
                "figsize": figsize,
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(1, 1, **kwargs_subplot)

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        # add horizontal lines
        axes.axhline(
            y=self.soil_properties.groundwater_level_ref,
            color="tab:blue",
            linestyle="--",
            label="Groundwater level",
        )
        axes.axhline(
            y=self.soil_properties.surface_level_ref,
            color="tab:brown",
            linestyle="--",
            label="Surface level",
        )

        # add bearing result subplot
        axes.plot(
            np.array(self.table.F_nk_d),
            self.table.pile_tip_level_nap,
            color="tab:orange",
            label="Fnk;d",
        )
        axes.plot(
            np.array(self.table.R_c_d_net),
            self.table.pile_tip_level_nap,
            label=r"Rc;net;d",
            lw=3,
            color="tab:blue",
        )
        axes.set_xlabel("Force [kN]")

        # add legend
        if add_legend:
            axes.legend(
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )

        # set grid
        axes.grid()

        return axes

    def plot_bearing_overview(
        self,
        figsize: Tuple[float, float] = (10.0, 12.0),
        width_ratios: Tuple[float, float, float] = (1, 0.1, 2),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Figure:
        """
        Plot an overview of the bearing-capacities, including the .

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        width_ratios:
            Tuple of width-ratios of the subplots, as the `plt.GridSpec` argument.
        add_legend:
            Add a legend to the second axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        fig:
            The matplotlib Figure
        """

        kwargs_subplot = {
            "gridspec_kw": {"width_ratios": width_ratios},
            "sharey": "row",
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        fig, _ = plt.subplots(
            1,
            3,
            **kwargs_subplot,
        )

        ax_qc, ax_layers, ax_bearing = fig.axes
        ax_rf = ax_qc.twiny()
        assert isinstance(ax_rf, Axes)

        # Plot bearing capacities
        self.soil_properties.cpt_table.plot_qc(ax_qc, add_legend=False)
        self.soil_properties.cpt_table.plot_friction_ratio(ax_rf, add_legend=False)
        self.soil_properties.plot_layers(ax_layers, add_legend=False)
        self.plot_bearing_capacities(axes=ax_bearing, add_legend=False)

        if add_legend:
            ax_qc_legend_handles_list = ax_qc.get_legend_handles_labels()[0]
            ax_rf_legend_handles_list = ax_rf.get_legend_handles_labels()[0]
            ax_layers_legend_handles_list = get_soil_layer_handles()

            # Omit last 2 duplicate "bearing" handles
            # (groundwater_level and surface_level):
            ax_bearing_legend_handles_list = ax_bearing.get_legend_handles_labels()[0][
                2:
            ]

            handles_list = [
                *ax_qc_legend_handles_list,
                *ax_rf_legend_handles_list,
                *ax_layers_legend_handles_list,
                *ax_bearing_legend_handles_list,
            ]

            ax_bearing.legend(
                handles=handles_list,
                loc="upper left",
                bbox_to_anchor=(1, 1),
                title="name: " + self.soil_properties.test_id
                if self.soil_properties.test_id is not None
                else "name: unknown",
            )

        return fig


class MaxBearingResults:
    """Object containing the results for the maximum net design bearing capacity (R_c_d_net) for every CPT."""

    def __init__(self, cpt_results_dict: Dict[str, MaxBearingResult]):
        """
        Object containing the results for the maximum net design bearing capacity (R_c_d_net) for every CPT.

        Parameters
        ----------
        cpt_results_dict
            The results for the maximum net design bearing capacity (R_c_d_net) for every CPT.
        """
        self._cpt_results_dict = cpt_results_dict

    def __getitem__(self, test_id: str) -> MaxBearingResult:
        if not isinstance(test_id, str):
            raise TypeError(f"Expected a test-id as a string, but got: {type(test_id)}")

        return self.get_cpt_results(test_id)

    @property
    def cpt_results_dict(self) -> Dict[str, MaxBearingResult]:
        """The dictionary with the MaxBearingResult for each CPT."""
        return self._cpt_results_dict

    @property
    def test_ids(self) -> List[str]:
        """The test-ids of the CPTs."""
        return list(self.cpt_results_dict.keys())

    @property
    def results(self) -> List[MaxBearingResult]:
        """The computed results, as a list of MaxBearingResult objects."""
        return list(self.cpt_results_dict.values())

    def get_cpt_results(self, test_id: str) -> MaxBearingResult:
        """
        Returns the `MaxBearingResult` object for the provided test_id.
        """

        if test_id not in self.cpt_results_dict.keys():
            raise ValueError(
                f"No Cpt-results were calculated for this test-id: {test_id}. "
                "Please check the spelling or run a new calculation for this CPT."
            )

        return self.cpt_results_dict[test_id]

    def get_results_per_cpt(self, column_name: str) -> pd.DataFrame:
        """
        Returns a pandas dataframe with a single result-item, organized per CPT
        (test-id) and pile-tip-level-nap.

        Parameters
        ----------
        column_name:
            The name of the result-item / column name of the single-cpt-results table.
        """
        if column_name not in self.to_pandas().columns or column_name in [
            "pile_tip_level_nap",
            "test_id",
        ]:
            raise ValueError("Invalid column_name provided.")

        results = pd.pivot(
            self.to_pandas(),
            values=column_name,
            index="pile_tip_level_nap",
            columns="test_id",
        )
        return results.sort_values("pile_tip_level_nap", ascending=False)

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """Returns a total overview of all single-cpt results in a pandas.DataFrame representation."""
        df_list: List[pd.DataFrame] = []

        for test_id in self.cpt_results_dict:
            df = self.cpt_results_dict[test_id].table.to_pandas()
            df = df.assign(test_id=test_id)
            df = df.assign(x=self.cpt_results_dict[test_id].soil_properties.x)
            df = df.assign(y=self.cpt_results_dict[test_id].soil_properties.y)
            df_list.append(df)

        cpt_results_df = pd.concat(df_list)
        cpt_results_df = cpt_results_df.assign(
            pile_tip_level_nap=cpt_results_df.pile_tip_level_nap.round(2)
        )

        return cpt_results_df

    @lru_cache()
    def triangulation(self, pile_tip_level_nap: float) -> List[Dict[str, list]]:
        """
        Delaunay tessellation based on the CPT location

        Returns
        -------
        collection: List
            A list of dictionaries containing the tessellation
            geometry and corresponding cpt names:

                - geometry: List[Tuple[float, float]]
                - test_id: List[str]

        """
        _lookup = {
            (point.soil_properties.x, point.soil_properties.y): key
            for key, point in self.cpt_results_dict.items()
        }
        # select point with valid bearing capacity at pile tip level
        _points = (
            self.to_pandas()
            .loc[
                (self.to_pandas()["pile_tip_level_nap"] == pile_tip_level_nap)
                & (~pd.isna(self.to_pandas()["R_c_d_net"])),
                ["x", "y"],
            ]
            .to_numpy()
            .tolist()
        )

        # check if enough points Delaunay
        if len(_points) < 4:
            raise ValueError(
                "Not enough points at this pile tip level to construct "
                "the delaunay tessellation based on the CPT location."
            )
        tri = Delaunay(
            _points,
            incremental=False,
            furthest_site=False,
            qhull_options="Qbb",
        )
        geometries = np.array(_points)[tri.simplices]

        return [
            {
                "geometry": geometry.tolist(),
                "test_id": [_lookup[(xy[0], xy[1])] for xy in geometry],
            }
            for geometry in geometries
        ]

    def plot(
        self,
        projection: Optional[Literal["3d"]] = "3d",
        hue: Literal["colormap", "category"] = "colormap",
        pile_load_uls: float = 100,
        figsize: Tuple[int, int] | None = None,
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plot a 3D scatterplot of the valid ULS load.

        Parameters
        ----------
        projection
            default is 3d
            The projection type of the subplot. use None to create a 2D plot
        hue
            default is colormap
            The marker colors methode. If colormap is used the colors represent the `R_c_d_net` value.
            The category option sets the colors to valid ULS loads. Please use the pile_load_uls attribute to set
            the required bearing capacity.
        pile_load_uls
            default is 100 kN
            ULS load in kN. Used to determine if a pile tip level configuration is valid.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        figure:
            The `Figure` object where the data was plotted on.
        """
        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)
        fig = plt.figure(**kwargs_subplot)
        axes = fig.add_subplot(projection=projection)
        df = self.to_pandas().dropna()
        # create color list based on hue option
        if hue == "category":
            colors = [
                "red" if var < pile_load_uls else "green" for var in df["R_c_d_net"]
            ]
        else:
            colors = df["R_c_d_net"].tolist()
        # create scatter plot
        if projection == "3d":
            cmap = axes.scatter(
                df["x"],
                df["y"],
                df["pile_tip_level_nap"],
                c=colors,
            )
            axes.set_xlabel("X")
            axes.set_ylabel("Y")
            axes.set_zlabel("Z [m w.r.t NAP]")

            # set cpt names
            for key, result in self.cpt_results_dict.items():
                axes.text(
                    result.soil_properties.x,
                    result.soil_properties.y,
                    result.table.pile_tip_level_nap.max(),
                    key,
                    "z",
                )
        else:
            cmap = axes.scatter(
                df["test_id"],
                df["pile_tip_level_nap"],
                c=colors,
            )
            axes.set_ylabel("Z [m w.r.t NAP]")
            axes.tick_params(axis="x", labelrotation=90)
            axes.grid()

        if hue == "category":
            fig.legend(
                title="$R_{c;d;net}$ [kN]",
                title_fontsize=18,
                fontsize=15,
                loc="lower right",
                handles=[
                    patches.Patch(
                        facecolor=color,
                        label=label,
                        alpha=0.9,
                        linewidth=2,
                        edgecolor="black",
                    )
                    for label, color in zip(
                        [f">= {pile_load_uls}", f"< {pile_load_uls}"],
                        ["green", "red"],
                    )
                ],
            )
        else:
            fig.colorbar(cmap, orientation="vertical", label="$R_{c;d;net}$ [kN]")

        return fig

    def map(
        self,
        pile_tip_level_nap: float,
        pile_load_uls: float = 100,
        show_delaunay_vertices: bool = True,
        show_voronoi_vertices: bool = False,
        figsize: Tuple[int, int] | None = None,
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plot a map of the valid ULS load for a given depth.

        Note
        ------
        Based on the Delaunay methode a tessellation is created with
        the location of the CPT's. Each triangle is then colored according to
        the bearing capacity of the CPT its based on. If any of the CPT does
        not meet the required capacity the triangle becomes also invalid.

        Warnings
        --------
        Please note that this map indication of valid ULS zones is intended as a visual aid to help
        the geotechnical engineer. It does not necessarily comply with the NEN 9997-1+C2:2017 since the NEN is open
        to interpretation. It is therefore that the interpretation provided by this methode must be carefully
        validated by a geotechnical engineer.

        Parameters
        ----------
        pile_tip_level_nap:
            Pile tip level to generate map.
        pile_load_uls
            default is 100 kN
            ULS load in kN. Used to determine if a pile tip level configuration is valid.
        show_delaunay_vertices
            default is True
            Add delaunay vertices to the figure
        show_voronoi_vertices
            default is False
            Add voronoi vertices to the figure
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        figure:
            The `Figure` object where the data was plotted on.
        """
        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)
        fig, axes = plt.subplots(**kwargs_subplot)

        # filter data
        df = (
            self.to_pandas()
            .loc[self.to_pandas()["pile_tip_level_nap"] == pile_tip_level_nap]
            .dropna()
        )

        if df.empty:
            raise ValueError(
                "Pile tip level is not valid pile tip level. "
                "Please select one of the following pile tip level: "
                f"[{(self.to_pandas()['pile_tip_level_nap']).unique()}]"
            )

        df["valid"] = [
            False if var < pile_load_uls else True for var in df["R_c_d_net"]
        ]

        # iterate over geometry
        if show_delaunay_vertices:
            _patches = []
            for tri in self.triangulation(pile_tip_level_nap):
                color = (
                    "green"
                    if all(
                        df.where(df["test_id"].isin(tri["test_id"])).dropna()["valid"]
                    )
                    else "red"
                )
                _patches.append(
                    patches.Polygon(
                        np.array(tri["geometry"]), facecolor=color, edgecolor="grey"
                    )
                )

            collection = PatchCollection(_patches, match_original=True)
            axes.add_collection(collection)

        if show_voronoi_vertices:
            points = [
                (point.soil_properties.x, point.soil_properties.y)
                for point in self.cpt_results_dict.values()
            ]
            vor = Voronoi(points)
            voronoi_plot_2d(
                vor,
                show_vertices=False,
                show_points=False,
                ax=axes,
                line_colors="black",
                line_alpha=0.7,
                line_width=0.1,
                point_size=0.0,
            )

        # add the cpt names
        axes.scatter(
            df["x"],
            df["y"],
            c=["green" if val else "red" for val in df["valid"]],
        )
        for label, x, y in zip(df["test_id"], df["x"], df["y"]):
            axes.annotate(label, xy=(x, y), xytext=(3, 3), textcoords="offset points")
        axes.set_xlabel("X")
        axes.set_ylabel("Y")
        axes.ticklabel_format(useOffset=False)
        fig.legend(
            title="$R_{c;d;net}$ [kN]",
            title_fontsize=18,
            fontsize=15,
            loc="lower right",
            handles=[
                patches.Patch(
                    facecolor=color,
                    label=label,
                    alpha=0.9,
                    linewidth=2,
                    edgecolor="black",
                )
                for label, color in zip(
                    [f">= {pile_load_uls}", f"< {pile_load_uls}"],
                    ["green", "red"],
                )
            ],
        )
        axes.set_title(f"Pile tip level at: {pile_tip_level_nap} [m w.r.t NAP]")

        return fig



from pypilecore.viewers.viewer_cpt_group_results import ViewerCptGroupResults
from pypilecore.viewers.viewer_cpt_results import ViewerCptResults
from pypilecore.viewers.viewer_cpt_results_plan_view import ViewerCptResultsPlanView

__all__ = [
    "ViewerCptResults",
    "ViewerCptResultsPlanView",
    "ViewerCptGroupResults",
]



from __future__ import annotations  # noqa: F404

from typing import Any

import pandas as pd
from IPython.display import DisplayHandle, display
from ipywidgets import widgets
from natsort import natsorted

from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults
from pypilecore.viewers.interactive_figures.figure_cpt_results_plan_view import (
    FigureCPTResultsPlanView,
)


class ViewerCptResultsPlanView:
    """
    Viewer for the CPT results of the bearing capacity calculations in
    plan view for a fixed pile tip level (PTL).

    It offers the following layout:
        - Dropdown widgets:
            - Case: to select the case to show.
            - Result: to select the result to show.
            - Pile tip level: to select the pile tip level to show.
        - Figure CPT results vs. pile tip level:
            - X axis: X coordinate.
            - Y axis: Y coordinate.
            - Each point represents a different CPT, but the same pile tip level.
    """

    def __init__(self, cases_multi_results: CasesMultiCPTBearingResults) -> None:
        """Initialize the viewer.

        Parameters
        ----------
        cases_multi_results : CasesMultiCPTBearingResults
            The results of the bearing capacity calculations.

        Raises
        ------
        TypeError
            If 'cases_multi_results' are not of type 'CasesMultiCPTBearingResults'.
        """

        # Initialize figure CPT results in plan view
        self._figure_plan_view = FigureCPTResultsPlanView(
            cases_multi_results=cases_multi_results
        )

        # Set up control widgets
        self._case_dropdown = widgets.Dropdown(
            description="Case:",
            value=self._figure_plan_view.cases[0],
            options=self._figure_plan_view.cases,
        )
        _options = natsorted(
            pd.unique(cases_multi_results.cpt_results_dataframe.result_name)
        )
        self._result_dropdown = widgets.Dropdown(
            description="Result:",
            value=_options[0],
            options=_options,
        )
        self._pile_tip_level_dropdown = widgets.Dropdown(
            description="Pile tip level NAP:",
            value=self._figure_plan_view.pile_tip_levels_nap[-1],
            options=self._figure_plan_view.pile_tip_levels_nap,
        )

        # Update plot for initial selection of control widgets
        self._update_case_result_and_ptl(None)

        # Set up callbacks
        self._case_dropdown.observe(self._update_case_result_and_ptl, "value")
        self._result_dropdown.observe(self._update_case_result_and_ptl, "value")
        self._pile_tip_level_dropdown.observe(self._update_case_result_and_ptl, "value")

        # Set up layout
        self._control_widgets = widgets.HBox(
            [
                self._case_dropdown,
                self._result_dropdown,
                self._pile_tip_level_dropdown,
            ]
        )
        self._layout = widgets.VBox(
            [self._control_widgets, self._figure_plan_view.figure]
        )

    def _update_case_result_and_ptl(self, change: Any) -> None:
        """
        Private method to update the figure when the case, result name or pile tip level
        are changed in the control widgets.
        """
        # Update the figure
        self._figure_plan_view.show_case_result_and_ptl(
            case_name=self._case_dropdown.value,
            result_name=self._result_dropdown.value,
            pile_tip_level_nap=self._pile_tip_level_dropdown.value,
        )

    def display(self) -> DisplayHandle | None:
        """Display the figure."""
        return display(self._layout)



from __future__ import annotations

import datetime
from copy import deepcopy
from typing import Any, Dict, List, Literal, Mapping, Sequence, Tuple

from pygef.cpt import CPTData

from pypilecore.common.piles import PileGridProperties, PileProperties
from pypilecore.input.soil_properties import create_soil_properties_payload


def create_multi_cpt_payload(
    pile_tip_levels_nap: Sequence[float],
    cptdata_objects: List[CPTData],
    classify_tables: Dict[str, dict],
    groundwater_level_nap: float,
    pile: PileProperties,
    pile_grid: PileGridProperties | None,
    stiff_construction: bool = False,
    ocr: float | None = None,
    soil_load_sls: float = 0.0,
    pile_head_level_nap: float | Literal["surface"] = "surface",
    excavation_depth_nap: float | None = None,
    top_of_tension_zone_nap: float | None = None,
    individual_top_of_tension_zone_nap: Mapping[Any, float] | None = None,
    excavation_param_t: float = 1.0,
    individual_ocr: Mapping[str, float] | None = None,
    overrule_xi: float | None = None,
    gamma_f_nk: float = 1.0,
    gamma_r_s: float = 1.2,
    gamma_r_b: float = 1.2,
    void_ratio_max: float = 0.8,
    void_ratio_min: float = 0.4,
    pile_load_sls_max: float = 1,
    pile_load_sls_min: float = 0,
    gamma_s_t: float = 1.35,
    gamma_gamma: float = 1.1,
    construction_sequence: Literal["cpt-pile", "pile-cpt"] = "cpt-pile",
) -> Tuple[dict, Dict[str, dict]]:
    """
    Creates a dictionary with the payload content for the PileCore endpoints
    "/tension/nen/multi-cpt/results" or "/tension/cur/multi-cpt/results"

    This dictionary can be passed directly to `nuclei.client.call_endpoint()`.

    Note
    ----
    The result should be converted to a jsonifyable message before it can be passed to a
    `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    pile_tip_levels_nap:
        Range of values for which the pile resistance is calculated.
    cptdata_objects:
        A list of pygef.CPTData objects
    classify_tables:
        A dictionary, mapping `CPTData.alias` values to dictionary with the resulting response
        of a call to CPTCore `classify/*` information, containing the following keys:

            - geotechnicalSoilName: Sequence[str]
                Geotechnical Soil Name related to the ISO
            - lowerBoundary: Sequence[float]
                Lower boundary of the layer [m]
            - upperBoundary: Sequence[float]
                Upper boundary of the layer [m]
            - color: Sequence[str]
                Hex color code
            - mainComponent: Sequence[Literal["rocks", "gravel", "sand", "silt", "clay", "peat"]]
                Main soil component
            - cohesion: Sequence[float]
                Cohesion of the layer [kPa]
            - gamma_sat: Sequence[float]
                Saturated unit weight [kN/m^3]
            - gamma_unsat: Sequence[float]
                Unsaturated unit weight [kN/m^3]
            - phi: Sequence[float]
                Phi [degrees]
            - undrainedShearStrength: Sequence[float]
                Undrained shear strength [kPa]

    groundwater_level_nap:
        The groundwater level. Unit: [m] w.r.t. NAP.
    pile:
        A PileProperties object.
    stiff_construction:
        Set to True if it's a stiff costruction. This will have influence on the xi factor
        if you don't overrule it. Default = False.
    ocr:
        The Over-Consolidation-Ratio [-] of the foundation layer.
    soil_load_sls:
        Load on soil surface [kPa], used to calculate soil settlement. This is only required
        with the settlement-driven friction-range strategy.
        Default = 0.0
    pile_head_level_nap:
        The level of the pile head. Can be:

            - float;
                This is interpreted as an absolute level in [m w.r.t. NAP];
            - Literal["surface"] (default).
                In this case, the soil_properties.service_level
                property is used.
    excavation_depth_nap:
        Soil excavation depth after the CPT was taken. Unit: [m] w.r.t. NAP.
    excavation_param_t:
        Excavation parameter depending on pile installation method and/or phasing.
            - Use 1.0 (default) for post-excavation installation with vibration (i.e. hammering).
            - Use 0.5 for reduced-vibration installation, or pile installation prior to
              excavation method.

        See for more info NEN 9997-1+C2:2017 7.6.2.3.(10)(k)
    overrule_xi:
        Set a fixed value for xi in all calculations. Use with caution. This will
        overrule the calculation of xi-values based on the group-size, variation-
        coefficient and construction stiffness.
    gamma_f_nk:
        Safety factor for design-values of the negative sleeve friction force.
        Default = 1.0
    gamma_r_s:
        Safety factor, used to obtain design-values of the pile-tip bearingcapacity.
        Default = 1.2
    gamma_r_b:
        Safety factor, used to obtain design-values of the sleeve bearingcapacity.
        Default = 1.2
    void_ratio_max:
        Maximum void ratio of the soil (the loosest packing). The influence of this
        parameter is limited and therefore it is typically sufficient to provide a
        global estimation. For normally consolidated sands in The Netherlands, an
        emax = 0.80 can be used in most cases.
        Default = 0.8
    void_ratio_min:
        Minimum void ratio of the soil (the densest packing). The influence of this parameter
        is limited and therefore it is typically sufficient to provide a global estimation.
        For normally consolidated sands in The Netherlands, an emin = 0.40 can be used in most cases.
        Default = 0.4
    pile_load_sls_max:
        Maximum tension force on pile Ft;max;k. Note that only positive values (tension force)
        are accepted. unit: kN
        Default = 1
    pile_load_sls_min:
        Minimum tension force (or maximum compression force) on pile -Ft;min;k (tension > 0). Positive
        values (tension force), negative values (compression force) and 0.0 are accepted. Note that
        the positive must be <= pile_load_sls_max. unit: kN
        Default = 0
    gamma_s_t:
        Pile resistance factor gamma_s;t used to compute the design cone resistance values
        qc;z;d as prescribed in NEN 9997-1+C2_2017 7.6.3.3(d).
        Default = 1.35
    gamma_gamma:
        Partial factor for volumetric weight NEN 9997-1+C2:2017 A.3.2
        Default = 1.1
    construction_sequence:
        Value that indicates if CPT are performed before or after pile
        installation according to 7.3.1 CUR 2001-4.
        Default = "cpt-pile"

        Notes:
            - If standard "NEN9997-1" is used the following applies:
                - `cpt-pile`, then the `f1` factor is computed according to NEN 9997-1+C2:2017 7.6.3.3(e).
                - `pile-cpt`, then the `f1` factor is 1.0 at all depths.
            - If standard "CUR236" is used `f1` is always 1.0 regardless of the construction sequence.

    Returns
    -------
    multi_cpt_payload:
        Dictionary with the payload content for the PileCore endpoint
        "/tension/nen/multi-cpt/results" or "/tension/cur/multi-cpt/results" # TODO: check if valid for both
    results_kwargs:
        Dictionary with keyword arguments for the `pilecore.MultiCPTTensionBearingResults`
        object.

    Raises
    ------
    ValueError:
        - If `excavation_depth_nap` is not None and `excavation_param_t` is None.
        - If both `relative_pile_load` and `pile_load_sls` are None.
    """
    # Input validation
    if excavation_depth_nap is not None and excavation_param_t is None:
        raise ValueError(
            "`excavation_param_t` cannot be None when `excavation_depth_nap` is not None."
        )

    soil_properties_list, results_kwargs = create_soil_properties_payload(
        cptdata_objects=cptdata_objects,
        classify_tables=classify_tables,
        groundwater_level_nap=groundwater_level_nap,
        friction_range_strategy="manual",
        excavation_depth_nap=excavation_depth_nap,
        master_ocr=ocr,
        individual_negative_friction_range_nap=None,
        individual_positive_friction_range_nap=None,
        individual_ocr=individual_ocr,
        master_top_of_tension_zone_nap=top_of_tension_zone_nap,
        individual_top_of_tension_zone_nap=individual_top_of_tension_zone_nap,
    )
    multi_cpt_payload: dict = dict(
        pile_tip_levels_nap=list(pile_tip_levels_nap),
        list_soil_properties=soil_properties_list,
        pile_properties=pile.serialize_payload(),
        pile_head_level_nap=pile_head_level_nap,
        stiff_construction=stiff_construction,
        soil_load=soil_load_sls if soil_load_sls is not None else 0.0,
        excavation_param_t=excavation_param_t,
        gamma_f_nk=gamma_f_nk,
        gamma_r_b=gamma_r_b,
        gamma_r_s=gamma_r_s,
        void_ratio_max=void_ratio_max,
        void_ratio_min=void_ratio_min,
        pile_load_sls_max=pile_load_sls_max,
        pile_load_sls_min=pile_load_sls_min,
        gamma_s_t=gamma_s_t,
        gamma_gamma=gamma_gamma,
        construction_sequence=construction_sequence,
    )

    # Add optional properties
    if excavation_depth_nap is not None:
        multi_cpt_payload["excavation_depth_nap"] = excavation_depth_nap
    if overrule_xi is not None:
        multi_cpt_payload["overrule_xi"] = overrule_xi
    if pile_grid is not None:
        multi_cpt_payload["pile_grid"] = pile_grid.serialize_payload()

    return multi_cpt_payload, results_kwargs


def create_multi_cpt_report_payload(
    multi_cpt_payload: dict,
    project_name: str,
    project_id: str,
    author: str,
    date: str | None = None,
    group_results_content: bool = True,
    individual_cpt_results_content: bool = True,
    result_summary_content: bool = True,
) -> dict:
    """
    Creates a dictionary with the payload content for the PileCore endpoint
    "/compression/multi-cpt/report"

    This dictionary can be passed directly to `nuclei.client.call_endpoint()`. Note that
    it should be converted to a jsonifyable message before it can be passed to a
    `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    multi_cpt_payload:
        Index 0 of the resulting tuple of a call to `create_multi_cpt_payload()`
    project_name:
        The name of the project.
    project_id:
        The identifier (code) of the project.
    author:
        The author of the report.
    date:
        The date of the report. If None, the current date will be used.
    group_results_content:
        Whether or not to add the group-results section to the report. Default = True
    individual_cpt_results_content:
        Whether or not to add the individual-cpt-results section to the report.
        Default = True
    result_summary_content:
        Whether or not to add the result-summary section to the report. Default = True

    Returns
    -------
    report_payload:
        Dictionary with the payload content for the PileCore endpoint
        "/compression/multi-cpt/report"
    """
    report_payload = deepcopy(multi_cpt_payload)
    report_payload.update(
        dict(
            content=dict(
                group_results=group_results_content,
                individual_cpt_results=individual_cpt_results_content,
                result_summary=result_summary_content,
            ),
            general=dict(
                author=author,
                project_id=project_id,
                project_name=project_name,
                date=(
                    date if date is None else datetime.date.today().strftime("%d-%m-%y")
                ),
            ),
        )
    )
    return report_payload



from .components import (
    PrimaryPileComponentDimension,
    RectPileGeometryComponent,
    RoundPileGeometryComponent,
    _BasePileGeometryComponent,
)
from .main import PileGeometry
from .materials import Color, PileMaterial

__all__ = [
    "PileGeometry",
    "PileMaterial",
    "Color",
    "PrimaryPileComponentDimension",
    "RoundPileGeometryComponent",
    "RectPileGeometryComponent",
    "_BasePileGeometryComponent",
]



from .common import PrimaryPileComponentDimension, _BasePileGeometryComponent
from .rectangle import RectPileGeometryComponent
from .round import RoundPileGeometryComponent

__all__ = [
    "_BasePileGeometryComponent",
    "PrimaryPileComponentDimension",
    "RectPileGeometryComponent",
    "RoundPileGeometryComponent",
]



from __future__ import annotations

from typing import Dict, List, Literal

from pypilecore.common.piles.geometry import PileGeometry
from pypilecore.common.piles.geometry.components import (
    PrimaryPileComponentDimension,
    RectPileGeometryComponent,
    RoundPileGeometryComponent,
)
from pypilecore.common.piles.geometry.materials import MATERIALS, PileMaterial
from pypilecore.common.piles.type import PileType


class PileProperties:
    """The PileProperties class represents all properties of a pile."""

    def __init__(
        self, geometry: PileGeometry, pile_type: PileType, name: str | None = None
    ):
        """
        Represents all properties of a pile.

        Parameters:
        -----------
        geometry: PileGeometry
            The geometry of the pile.
        pile_type: PileType
            The type of the pile.
        name: str, optional
            The name of the pile, by default None.
        """
        self._geometry = geometry
        self._pile_type = pile_type
        self._name = name

    @classmethod
    def from_api_response(cls, pile_properties: dict) -> PileProperties:
        """
        Instantiates a PileProperties from a response dictionary.

        Parameters:
        -----------
        pile_properties: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties".

        Returns:
        --------
        PileProperties
            A pile properties object.
        """
        return cls(
            geometry=PileGeometry.from_api_response(pile_properties["geometry"]),
            pile_type=PileType.from_api_response(pile_properties["pile_type"]),
            name=pile_properties.get("name"),
        )

    @property
    def geometry(self) -> PileGeometry:
        """The geometry of the pile."""
        return self._geometry

    @property
    def pile_type(self) -> PileType:
        """The type of the pile."""
        return self._pile_type

    @property
    def name(self) -> str | None:
        """The name of the pile."""
        if self._name is not None:
            return str(self._name)
        return None

    def serialize_payload(self) -> Dict[str, dict | str]:
        """
        Serialize the pile properties to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the geometry, pile type, and name (if set).
        """
        payload: Dict[str, dict | str] = {
            "geometry": self.geometry.serialize_payload(),
            "pile_type": self.pile_type.serialize_payload(),
        }

        if self.name is not None:
            payload["name"] = self.name

        return payload


def create_basic_pile(
    pile_shape: Literal["round", "rect", "rectangle"],
    pile_name: str | None = None,
    main_type: Literal["concrete", "steel", "wood", "anchor"] | None = None,
    specification: str | int | float | None = None,
    installation: str | None = None,
    height_base: float | None = None,
    core_secondary_dimension: float | None = None,
    core_tertiary_dimension: float | None = None,
    base_secondary_dimension: float | None = None,
    base_tertiary_dimension: float | None = None,
    core_diameter: float | None = None,
    base_diameter: float | None = None,
    pile_material: Literal["concrete", "steel", "wood", "grout", "grout_extorted"]
    | None = None,
    custom_material: dict | None = None,
    settlement_curve: int | None = None,
    adhesion: float | None = None,
    alpha_p: float | None = None,
    alpha_s_clay: float | None = None,
    alpha_s_sand: float | None = None,
    beta_p: float | None = None,
    alpha_t_clay: float | None = None,
    alpha_t_sand: float | None = None,
    pile_tip_factor_s: float | None = None,
    is_auger: bool | None = None,
    is_low_vibrating: bool | None = None,
    negative_fr_delta_factor: float | None = None,
    qc_z_a_lesser_1m: float | None = None,
    qc_z_a_greater_1m: float | None = None,
    qb_max_limit: float | None = None,
    chamfered: float | None = None,
) -> PileProperties:
    """
    Create a basic pile with the provided parameters.
    A "basic" pile is a pile with a basic geometry, containing at most 2 components,
    namely a core component and an optional shell/widened-base component.

    Parameters
    ----------
    pile_shape :
        The shape of the pile.
        Both "rect" and "rectangle" are accepted for rectangular piles.
    pile_name :
        The name of the pile, by default None.
    main_type :
        The main type of the standard pile type definition, by default None.
    specification :
        The specification of the standard pile type definition, by default None.
        Required if `main_type` is provided.
    installation :
        The installation of the standard pile type definition, by default None.
    height_base :
        The height [m] of the base component, by default None.
        Required if base dimensions are provided.
    core_secondary_dimension :
        The largest dimension [m] of the rectangular core component's cross-section, by default None.
        Required for rectangular piles.
    core_tertiary_dimension :
        The smallest dimension [m] of the rectangular core component's cross-section, by default None.
        If None and pile shape is rectangular, `core_tertiary_dimension` is used resulting in a square core.
    base_secondary_dimension :
        The largest dimension [m] of the rectangular base/shell component's cross-section, by default None.
        If None and pile shape is rectangular, then the base has the same dimensions as the core.
        Requires `height_base` to be provided.
    base_tertiary_dimension :
        The smallest dimension [m] of the rectangular base/shell component's cross-section, by default None.
        If None and pile shape is rectangular, then:
            - If `base_secondary_dimension` is None, then the base has the same dimensions as the core.
            - If `base_secondary_dimension` is provided, then `base_tertiary_dimension` is used resulting
            in a square base.
        Requires `height_base` to be provided.
    core_diameter :
        The diameter [m] of the round core component's cross-section, by default None.
        Required for round piles.
    base_diameter :
        The diameter [m] of the round base/shell component's cross-section, by default None.
        If None and pile shape is round, then the base has the same diameter as the core.
        Requires `height_base` to be provided.
    pile_material :
        The material of the pile, by default None.
        It is ignored if `custom_material` is provided.
        Otherwise:
            - If None, it is inferred from the standard pile.
            - Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type and no custom material were specified.
    custom_material :
        A custom material for the pile, by default None.
        If provided, it is used even if a standard pile was selected.
        Required keys and value types:
            - "name": str
            - "elastic_modulus: float
                Unit: [MPa]
            - "color": str
                Hexadecimal color
        Example: {
            "name": "custom_material",
            "elastic_modulus": 15e3,
            "color": "#ff0000",
        }
    settlement_curve :
        The settlement curve of the pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    adhesion :
        The adhesion value [kPa] of the pile, by default None.
        Use it if the pile shaft has undergone a special treatment.
        If None, then adhesion is disregarded.
        Examples:
            - adhesion = 50 kPa for synthetic coating
            - adhesion = 20 kPa for bentonite
            - adhesion = 10 kPa for bitumen coating
    alpha_p :
        The alpha_p of the pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    alpha_s_clay :
        The alpha_s_clay of the pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    alpha_s_sand :
        The alpha_s_sand of the pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    alpha_t_clay :
        The alpha_t_clay of the pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    alpha_t_sand :
        The alpha_t_sand of the pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    beta_p :
        The beta_p of the pile, by default None.
        If None, the value is computed from the pile dimensions.
    pile_tip_factor_s :
        The pile tip factor s of the pile, by default None.
        If None, the value is computed from the pile dimensions and soil properties.
    is_auger :
        Whether the pile is an auger pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    is_low_vibrating :
        Whether the pile is low vibrating, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    negative_fr_delta_factor :
        The negative friction delta factor of the pile, by default None.
        If None, it is inferred from the standard pile.
        Overwrites the value of the standard pile if provided together with a standard pile.
        Required if no standard pile type was specified.
    qc_z_a_lesser_1m :
        Maximum cone resistance qc value [MPa] allowed for layers with thickness < 1m in the calculation of positive skin friction resistance.
        It must be less or equal to `qc_z_a_greater_1m`.
        If None, then 12 MPa is used.
        Default: None.
    qc_z_a_greater_1m :
        Maximum cone resistance qc value [MPa] allowed for layers with thickness >= 1m in the calculation of positive skin friction resistance.
        It must be greater or equal to `qc_z_a_lesser_1m`.
        If None, then 15 MPa is used.
        Default: None.
    qb_max_limit :
        Maximum value allowed for the pile tip resistance qb_max [MPa].
        Note that is value will be used as a limit for qb_max unless use_almere_rules is true and a higher value than 12 MPa is specified.
        If None, then 15 MPa is used.
        Default: None.
    chamfered :
        The chamfered value of the pile type, by default None.
        Required if no standard pile type was specified.

    Returns
    -------
    PileProperties
        The pile properties object.
    """
    materials = [
        PileMaterial(**material_value) for material_value in MATERIALS.values()  # type: ignore
    ]

    if custom_material is not None:
        materials = [PileMaterial(**custom_material)]

    if pile_shape == "round":
        if core_diameter is None:
            raise ValueError("`core_diameter` must be provided for a round pile.")

        components: List[RoundPileGeometryComponent | RectPileGeometryComponent] = [
            RoundPileGeometryComponent(
                diameter=core_diameter,
                primary_dimension=PrimaryPileComponentDimension(
                    length=None,
                ),
                material=pile_material,
            )
        ]

        if base_diameter is not None:
            components.append(
                RoundPileGeometryComponent(
                    diameter=base_diameter,
                    primary_dimension=PrimaryPileComponentDimension(
                        length=height_base,
                    ),
                    material=pile_material,
                )
            )

    elif pile_shape in ["rect", "rectangle"]:
        if core_secondary_dimension is None:
            raise ValueError(
                "`core_secondary_dimension` must be provided for a rectangular pile."
            )

        components = [
            RectPileGeometryComponent(
                primary_dimension=PrimaryPileComponentDimension(length=None),
                secondary_dimension=core_secondary_dimension,
                tertiary_dimension=core_tertiary_dimension,
                material=pile_material,
            )
        ]

        if base_secondary_dimension is not None:
            components.append(
                RectPileGeometryComponent(
                    secondary_dimension=base_secondary_dimension,
                    tertiary_dimension=base_tertiary_dimension,
                    primary_dimension=PrimaryPileComponentDimension(
                        length=height_base,
                    ),
                    material=pile_material,
                )
            )

    else:
        raise ValueError(f"Invalid pile shape: {pile_shape}")

    geometry = PileGeometry(
        components=components,
        materials=materials,
        pile_tip_factor_s=pile_tip_factor_s,
        beta_p=beta_p,
    )

    if main_type is not None:
        if specification is None:
            raise ValueError("Specification must be provided if main type is provided.")

        standard_pile: Dict[str, str | int] | None = {
            "main_type": main_type,
            "specification": int(specification),
        }

        if installation is not None:
            standard_pile["installation"] = installation  # type: ignore

    else:
        standard_pile = None

    pile_type = PileType(
        standard_pile=standard_pile,
        alpha_s_sand=alpha_s_sand,
        alpha_s_clay=alpha_s_clay,
        alpha_p=alpha_p,
        alpha_t_sand=alpha_t_sand,
        alpha_t_clay=alpha_t_clay,
        settlement_curve=settlement_curve,
        negative_fr_delta_factor=negative_fr_delta_factor,
        adhesion=adhesion,
        is_low_vibrating=is_low_vibrating,
        is_auger=is_auger,
        qc_z_a_lesser_1m=qc_z_a_lesser_1m,
        qc_z_a_greater_1m=qc_z_a_greater_1m,
        qb_max_limit=qb_max_limit,
        chamfered=chamfered,
    )

    return PileProperties(geometry=geometry, pile_type=pile_type, name=pile_name)



from __future__ import annotations

from typing import Any, Dict, List, Tuple

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy.typing import NDArray

from pypilecore.common.piles.geometry.components import (
    RectPileGeometryComponent,
    RoundPileGeometryComponent,
)
from pypilecore.common.piles.geometry.materials import Color, PileMaterial


class PileGeometry:
    """The PileGeometry class represents the geometry of a pile."""

    def __init__(
        self,
        components: List[RoundPileGeometryComponent | RectPileGeometryComponent],
        materials: List[PileMaterial] | None = None,
        pile_tip_factor_s: float | None = None,
        beta_p: float | None = None,
    ):
        """
        Represents the geometry of a pile.

        Parameters:
        -----------
        components : list
            A list of pile geometry components.
        materials : list, optional
            A list of materials used in the pile geometry, by default None.
        pile_tip_factor_s : float, optional
            The pile tip factor S, by default None.
        beta_p : float, optional
            The beta_p value, by default None.
        """
        self._components = components
        self._materials = materials
        self._pile_tip_factor_s = pile_tip_factor_s
        self._beta_p = beta_p

    @classmethod
    def from_api_response(cls, geometry: dict) -> PileGeometry:
        """
        Instantiates a PileGeometry from a geometry object in the API response payload.

        Parameters:
        -----------
        geometry: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties/geometry".

        Returns:
        --------
        PileGeometry
            A pile geometry.
        """
        components: List[RoundPileGeometryComponent | RectPileGeometryComponent] = []
        for component in geometry["components"]:
            if component["outer_shape"] == "round":
                components.append(
                    RoundPileGeometryComponent.from_api_response(component)
                )
            else:
                components.append(
                    RectPileGeometryComponent.from_api_response(component)
                )

        materials = []
        if "materials" in geometry:
            for material in geometry["materials"]:
                materials.append(PileMaterial.from_api_response(material))

        return cls(
            components=components,
            materials=materials,
            pile_tip_factor_s=geometry["properties"].get("pile_tip_factor_s"),
            beta_p=geometry["properties"]["beta_p"],
        )

    @property
    def components(
        self,
    ) -> List[RoundPileGeometryComponent | RectPileGeometryComponent]:
        """The components of the pile geometry"""
        return self._components

    @property
    def materials(self) -> List[PileMaterial]:
        """The materials used in the pile geometry"""
        return self._materials if self._materials is not None else []

    @property
    def materials_dict(self) -> Dict[str, PileMaterial]:
        """The materials used in the pile geometry as a dictionary with the material name as key"""
        return {material.name: material for material in self.materials}

    @property
    def pile_tip_factor_s(self) -> float | None:
        """The pile tip factor S of the pile geometry"""
        return self._pile_tip_factor_s

    @property
    def beta_p(self) -> float | None:
        """The beta_p value of the pile geometry"""
        return self._beta_p

    @property
    def equiv_diameter_pile_tip(self) -> float:
        """The equivalent diameter of the pile at the pile tip."""
        return self.components[-1].equiv_tip_diameter

    @property
    def circumference_pile_tip(self) -> float:
        """The outer-circumference of the pile at the pile tip."""
        return self.components[-1].circumference

    @property
    def area_pile_tip(self) -> float:
        """The area of the pile at the pile tip."""
        return self.components[-1].area_full

    def serialize_payload(self) -> Dict[str, list | dict]:
        """
        Serialize the pile geometry to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the components, materials (if set), pile tip factor S (if set), and beta_p (if set).
        """
        components = [component.serialize_payload() for component in self.components]
        payload: Dict[str, Any] = {"components": components}

        if self.materials is not None and len(self.materials) > 0:
            materials = [material.serialize_payload() for material in self.materials]
            payload["materials"] = materials

        custom_geom_properties: Dict[str, float] = {}
        if self.pile_tip_factor_s is not None:
            custom_geom_properties["pile_tip_factor_s"] = self.pile_tip_factor_s

        if self.beta_p is not None:
            custom_geom_properties["beta_p"] = self.beta_p

        if len(custom_geom_properties.keys()) > 0:
            payload["custom_properties"] = custom_geom_properties

        return payload

    def get_circum_vs_depth(
        self,
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
        depth_nap: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns pile circumferences at requested depths.

        Parameters
        ---------_
        pile_tip_level_nap : float
            Pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            Pile head level in [m] w.r.t. NAP.
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with pile circumferences at the requested `depth_nap` levels.
        """
        circum_vs_depth = np.zeros_like(depth_nap)

        # Use the maximum circumference of all components at each depth.
        for component in self.components:
            circum_vs_depth = np.maximum(
                circum_vs_depth,
                component.get_circum_vs_depth(
                    pile_tip_level_nap=pile_tip_level_nap,
                    pile_head_level_nap=pile_head_level_nap,
                    depth_nap=depth_nap,
                ),
            )

        return circum_vs_depth

    def get_area_vs_depth(
        self,
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
        depth_nap: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns cross-sectional area of the pile at requested depths.

        Parameters
        ---------_
        pile_tip_level_nap : float
            Pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            Pile head level in [m] w.r.t. NAP.
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with pile areas at the requested `depth_nap` levels.
        """
        area_vs_depth = np.zeros_like(depth_nap)

        # Use the maximum area of all components at each depth.
        for component in self.components:
            area_vs_depth = np.maximum(
                area_vs_depth,
                component.get_area_vs_depth(
                    pile_tip_level_nap=pile_tip_level_nap,
                    pile_head_level_nap=pile_head_level_nap,
                    depth_nap=depth_nap,
                ),
            )

        return area_vs_depth

    def plot(
        self,
        pile_tip_level_nap: float | int = -10,
        pile_head_level_nap: float | int = 0,
        figsize: Tuple[float, float] | None = (3, 9),
        show: bool = True,
        **kwargs: Any,
    ) -> List[Axes]:
        """
        Plot the top-view of the pile at a specified depth.

        Parameters
        ----------
        pile_tip_level_nap : float, optional
            The pile tip level in m w.r.t. NAP, by default -10.
        pile_head_level_nap : float, optional
            The pile head level in m w.r.t. NAP, by default 0.
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        show : bool, optional
            Whether to display the plot, by default True.
        **kwargs
            Additional keyword arguments to pass to the `plt
        """
        kwargs_subplot = {
            "figsize": figsize,
            "gridspec_kw": {
                "hspace": 0.15,
                "height_ratios": [1, 4],
            },
        }

        kwargs_subplot.update(kwargs)

        height_ratio = (
            kwargs_subplot["gridspec_kw"]["height_ratios"][1]  # type: ignore
            / kwargs_subplot["gridspec_kw"]["height_ratios"][0]  # type: ignore
        )

        _, axes = plt.subplots(
            2,
            1,
            **kwargs_subplot,
        )

        x_ticks = set([0.0])
        y_ticks = set([0.0])

        for component in self.components[::-1]:
            facecolor = "grey"
            if component.material in self.materials_dict:
                material_color = self.materials_dict[component.material].color
                if isinstance(material_color, Color):
                    facecolor = material_color.hex
                elif isinstance(material_color, Color):
                    facecolor = (
                        material_color.red,
                        material_color.green,
                        material_color.blue,
                    )

            component.plot_cross_section_exterior(
                axes=axes[0],
                facecolor=facecolor,
                edgecolor="black",
                axis_arg=None,
                show=False,
            )

            component.plot_side_view(
                pile_tip_level_nap=pile_tip_level_nap,
                pile_head_level_nap=pile_head_level_nap,
                axes=axes[1],
                facecolor=facecolor,
                axis_arg=None,
                show=False,
            )

            x_ticks.add(component.cross_section_bounds[0])
            x_ticks.add(component.cross_section_bounds[1])
            y_ticks.add(component.cross_section_bounds[2])
            y_ticks.add(component.cross_section_bounds[3])
        axes[0].axis("scaled")
        axes[0].set(aspect=1)
        axes[1].axis("auto")
        ax1_aspect = (
            abs(axes[0].axis()[2] - axes[0].axis()[3])
            / abs(axes[1].axis()[2] - axes[1].axis()[3])
            * height_ratio
        )
        axes[1].set(aspect=ax1_aspect)

        axes[0].spines[:].set_visible(False)
        axes[1].spines[:].set_visible(False)

        axes[0].set_xticks(ticks=list(x_ticks))
        axes[0].set_yticks(ticks=list(y_ticks))
        axes[1].set_xticks(ticks=list(x_ticks))
        axes[1].set_yticks(
            ticks=[pile_head_level_nap, pile_tip_level_nap],
            labels=["Pile Head", "Pile Tip"],
        )
        axes[0].tick_params(axis="x", labelrotation=45)
        axes[1].tick_params(axis="x", labelrotation=45)

        if show:
            plt.show()
        return axes



from __future__ import annotations  # noqa: F404

from typing import Hashable, List

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults
from pypilecore.results.result_definitions import CPTResultDefinitions
from pypilecore.viewers.interactive_figures.utils import get_continuous_color


class FigureCPTResultsPlanView:
    """
    Interactive figure to show the CPT results of the bearing capacity calculations
    in plan view for a fixed pile tip level (PTL).

    The layout of the figure is:
        - X axis: X coordinate.
        - Y axis: Y coordinate.
        - Each point represents a different CPT, but the same pile tip level.

    The figure has a method to switch between case, result and pile tip level.
    """

    def __init__(self, cases_multi_results: CasesMultiCPTBearingResults) -> None:
        """
        Initializes the figure.

        Parameters
        ----------
        cases_multi_results : CasesMultiCPTBearingResults
            The results of the bearing capacity calculations.

        Raises
        ------
        TypeError
            If `cases_multi_results` are not of type 'CasesMultiCPTBearingResults'.
        """
        # Validate the data
        self._set_results(cases_multi_results)

        # Initialize the figure
        self._figure = go.FigureWidget()

    def _set_results(self, value: CasesMultiCPTBearingResults) -> None:
        """Private setter for the results."""
        if not isinstance(value, CasesMultiCPTBearingResults):
            raise TypeError(
                f"Expected type 'CasesMultiCPTBearingResults' for 'cases_multi_results', but got {type(value)}"
            )
        self._results = value

    @property
    def results(self) -> CasesMultiCPTBearingResults:
        """The results of the bearing capacity calculations."""
        return self._results

    @property
    def data(self) -> pd.DataFrame:
        """The dataframe used to plot the results."""
        return self.results.cpt_results_dataframe

    @property
    def cases(self) -> List[Hashable]:
        """The case names of each MultiCPTBearingResults."""
        return self.results.cases

    @property
    def test_ids(self) -> List[str]:
        """The test_ids (cpt names) of all the MultiCPTBearingResults."""
        return self.results.test_ids

    @property
    def pile_tip_levels_nap(self) -> List[float]:
        """The pile tip levels w.r.t. NAP of all the MultiCPTBearingResults."""
        return self.results.pile_tip_levels_nap

    @property
    def figure(self) -> go.FigureWidget:
        """The figure widget."""
        return self._figure

    def get_visible_test_ids(self) -> List[str]:
        """Returns the visible `test_id` (s) in the figure widget."""
        return [trace.name for trace in self.figure.data if trace.visible is True]

    def show_case_result_and_ptl(
        self, case_name: Hashable, result_name: str, pile_tip_level_nap: float
    ) -> None:
        """Shows the results for all CPTs for the requested `case_name`, `result_name` and `pile_tip_level`.

        Parameters
        ----------
        case_name : str
            The name of the case to show.
        result_name : str
            The name of the result to show.
        pile_tip_level_nap : float
            The pile tip level w.r.t. NAP to show.


        Raises
        ------
        TypeError
            If the `pile_tip_level_nap` is not of type 'float'.
        ValueError
            If the `case_name` is not found in the cases.
            If the `result_name` is not found in the CPTResultDefinitions.
            If the `pile_tip_level_nap` is not found in the pile tip levels.
        """
        # Check that case name is in cases.
        if case_name not in self.cases:
            raise ValueError(f"Case name '{case_name}' not found in cases.")

        # Get the result definition that corresponds to the result name.
        result_definition = CPTResultDefinitions.get(result_name)

        # Check that pile tip level NAP is in pile tip levels.
        if not isinstance(pile_tip_level_nap, (int, float)):
            raise TypeError(
                f"Expected type 'float' for 'pile_tip_level_nap', but got {type(pile_tip_level_nap)}"
            )
        if not any(
            np.isclose(pile_tip_level_nap, self.pile_tip_levels_nap, rtol=0, atol=1e-4)
        ):
            raise ValueError(
                f"Pile tip level NAP '{pile_tip_level_nap}' not found in pile tip levels NAP."
            )

        # Get the visible test_ids
        if len(self.figure.data) == 0:
            # Select all test_ids if there are no traces in the figure yet.
            visible_test_ids = self.test_ids
        else:
            visible_test_ids = self.get_visible_test_ids()

        # Select data for case name and result name.
        mask_case_name = (
            self.data["case_name"] == case_name
            if case_name is not None
            else self.data["case_name"].isna()
        )
        mask_pile_tip_level = np.isclose(
            pile_tip_level_nap,
            self.data["pile_tip_level_nap"].to_numpy(),
            rtol=0,
            atol=1e-4,
        )
        selected_data = self.data.loc[
            (mask_case_name)
            & (self.data["result_name"] == result_definition.name)
            & (mask_pile_tip_level)
        ]

        # Get the min and max result values for the color scale.
        result_max = selected_data["result"].max()
        result_min = selected_data["result"].min()
        colorscale = px.colors.get_colorscale("picnic")

        traces = []
        for test_id in self.test_ids:
            df = selected_data.loc[selected_data["test_id"] == test_id]
            result = 0 if df.empty else round(df["result"].values[0], 1)
            color = get_continuous_color(
                colorscale=colorscale,
                intermed=(result - result_min) / (result_max - result_min),
            )
            traces.append(
                go.Scatter(
                    x=df["x"],
                    y=df["y"],
                    text=f"CPT {test_id}<br>{result}",
                    mode="markers+text",
                    name=test_id,
                    marker=dict(
                        size=8,
                        color=color,
                        line=dict(
                            width=0.5,
                            color="black",
                        ),
                    ),
                    textposition="top center",
                    hoverinfo="none",
                )
            )

        # Add the colorbar
        traces.append(
            go.Scatter(
                x=[None],
                y=[None],
                mode="markers",
                name="__colorbar__",
                marker=dict(
                    colorscale=colorscale,
                    showscale=True,
                    cmin=result_min,
                    cmax=result_max,
                    colorbar=dict(
                        thickness=16,
                        orientation="h",
                        x=0.75,
                        y=1.0,
                        len=0.5,
                        title=f"{result_definition.value.html} [{result_definition.value.unit}]",
                        title_font_size=14,
                    ),
                ),
                hoverinfo="none",
                showlegend=False,
            )
        )

        with self.figure.batch_update():
            # Empty traces
            self.figure.data = []

            # Apply changes
            self.figure.add_traces(traces)

            self.figure.update_layout(
                title=f"CPT Results in Plan View<br>Case: {case_name}, "
                + f"Pile tip level [m NAP]: {pile_tip_level_nap}<br>"
                + f"Result: {result_definition.value.html} [{result_definition.value.unit}]",
                # The following parameters shouldn't be necessary to update
                # (and therefore could be assigned on initialization of self.figure),
                # but they are needed to avoid a bug in plotly since dash > 3.0.
                height=800,
                width=1200,
                legend_title="CPT",
                colorway=px.colors.qualitative.Plotly,
                xaxis=go.layout.XAxis(
                    title="X [m]",
                    title_font_size=18,
                ),
                yaxis=go.layout.YAxis(
                    title="Y [m]",
                    title_font_size=18,
                ),
                autosize=False,
            )

            self.figure.update_traces(
                selector=lambda x: x.name in visible_test_ids,
                patch=dict(
                    visible=True,
                ),
            )

            self.figure.update_traces(
                selector=lambda x: x.name not in visible_test_ids,
                patch=dict(
                    visible="legendonly",
                ),
            )

            self.figure.update_traces(
                selector=lambda x: x.name == "__colorbar__",
                patch=dict(
                    visible=True,
                    showlegend=False,
                ),
            )



from __future__ import annotations

from typing import Any, List, Tuple

import matplotlib.patches as patches
import matplotlib.ticker as ticker
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from shapely.geometry import MultiPolygon, shape
from shapely.plotting import plot_polygon


class PileGridProperties:
    def __init__(
        self,
        points: List[Tuple[float, float]],
        index_location: int,
        geometry: MultiPolygon | None = None,
        _validate: bool = True,
    ):
        """
        Class that holds the information of the pile placement.
        The structure of the grid effects the effective area of the
        pile.

        Parameters
        ----------
        points: List of tuple
            List of tuple with xy coordinates.
        index_location: int
            selected index of the points list as pile location
        geometry: MultiPolygon, optional
            A;eff as polygon
        _validate: bool, optional
            flag that overrules validation of the parameters

        """
        if len(points) < 2 and _validate:
            raise ValueError("Provide at least two points to the pile grid.")

        if len(points) != len(set(points)):
            raise ValueError("All points of the pile grid must be unique.")
        if len(points) - 1 < index_location > 0:
            raise IndexError(
                f"Index out of range, must be between 0 and {len(points) - 1} got {index_location}."
            )

        self._points = points
        self._index_location = index_location
        self._geometry = geometry

    @classmethod
    def regular(cls, ctc: float, index_location: int) -> "PileGridProperties":
        """
        Create a regular grid with a centre to centre distance provided by the user.
        The regular grid contains 9 points. Numbering start in the lower left corner.
        So for the
            - center pile `index_location= 4`
            - middle pile `index_location= 1 or 3 or 4 or 5 or 7`
            - corner pile `index_location= 0 or 2 or 6 or 8`

        x --- x --- x
        |     |     |
        x --- x --- x
        |     |     |
        x --- x --- x   with --- == | == ctc

        Parameters
        ----------
        ctc: float
            Centre to centre distance of regular grid [m]
        index_location: int
            Location of the pile to calculate
        Returns
        -------
        PileGrid
        """
        xx, yy = np.meshgrid(np.arange(0, 3) * ctc, np.arange(0, 3) * ctc)

        return cls(
            points=list(zip(xx.flatten(), yy.flatten())), index_location=index_location
        )

    @classmethod
    def from_api_response(cls, payload: dict) -> PileGridProperties:
        """
        Instantiates a PileGeometry from a geometry object in the API response payload.

        Parameters:
        -----------
        geometry: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties/geometry".

        Returns:
        --------
        PileGeometry
            A pile geometry.
        """
        # make list type hashable
        points = [(point[0], point[1]) for point in payload["points"]]
        return cls(
            points=points,
            index_location=payload["index_location"],
            geometry=shape(payload["geometry"]),
            _validate=False,
        )

    def serialize_payload(self) -> dict:
        return {
            "locations": self._points,
            "pile_index": self._index_location,
        }

    def plot_overview(
        self,
        figsize: Tuple[float, float] = (6.0, 6.0),
        axes: Axes | None = None,
        add_ticks: bool = False,
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the Bird's-eye view of the pile grid. When provided the effect area of
        the pile influence zone is add to the figure.

        Parameters
        ----------
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        axes : Axes
            The axes object to plot the cross-section on.
        add_ticks : bool
            Add ticks to figure, by default False
        add_legend : bool
            Add legend to figure, by default True
        **kwargs
            Additional keyword arguments to pass to the `plt
        """

        # Create axes objects if not provided
        if axes is not None:
            if not isinstance(axes, Axes):
                raise ValueError(
                    "'axes' argument to plot_overview() must be a `pyplot.axes.Axes` object or None."
                )
        else:
            kwargs_subplot = {
                "figsize": figsize,
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(1, 1, **kwargs_subplot)

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )
        handles = [
            Line2D(
                [0],
                [0],
                label="Selected pile",
                marker=".",
                color="None",
                markerfacecolor="tab:orange",
                markeredgecolor="None",
                ls="",
            ),
        ]

        if len(self._points) >= 2:
            handles.append(
                Line2D(
                    [0],
                    [0],
                    color="None",
                    label="Other piles",
                    marker=".",
                    markerfacecolor="gray",
                    markeredgecolor="None",
                    ls="",
                )
            )
        # add the effective area of the pile with the pile grid
        if self._geometry is not None:
            plot_polygon(self._geometry, ax=axes, add_points=False, color="tab:blue")
            handles.append(
                patches.Patch(
                    facecolor="tab:blue",
                    alpha=0.3,
                    label=r"$A_{eff}$" + rf" {self._geometry.area:.1f} $m^2$",
                    edgecolor="tab:blue",
                )
            )

        # plot points
        colors = ["gray"] * len(self._points)
        colors[self._index_location] = "tab:orange"
        axes.scatter(*zip(*self._points), color=colors, marker=".")

        # add labels to points
        for x, y, label in zip(*zip(*self._points), range(len(self._points))):
            axes.annotate(label, xy=(x, y), xytext=(3, 3), textcoords="offset points")

        axes.ticklabel_format(useOffset=False, style="plain")
        axes.set_aspect("equal", adjustable="box")
        if add_ticks:
            axes.xaxis.set_major_locator(ticker.NullLocator())
            axes.yaxis.set_major_locator(ticker.NullLocator())

        if add_legend:
            axes.legend(
                title="Pile grid configuration",
                bbox_to_anchor=(1.04, 1),
                loc="upper left",
                handles=handles,
            )
        return axes



from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults
from pypilecore.results.compression.multi_cpt_results import (
    MultiCPTCompressionBearingResults,
)
from pypilecore.results.compression.single_cpt_results import (
    SingleCPTCompressionBearingResults,
)
from pypilecore.results.grouper_result import GrouperResults
from pypilecore.results.soil_properties import SoilProperties
from pypilecore.results.tension.multi_cpt_results import MultiCPTTensionBearingResults
from pypilecore.results.tension.single_cpt_results import SingleCPTTensionBearingResults

__all__ = [
    "CasesMultiCPTBearingResults",
    "MultiCPTCompressionBearingResults",
    "SingleCPTCompressionBearingResults",
    "MultiCPTTensionBearingResults",
    "SingleCPTTensionBearingResults",
    "SoilProperties",
    "GrouperResults",
]



from pypilecore.viewers.interactive_figures.figure_cpt_group_results_versus_ptls import (
    FigureCPTGroupResultsVersusPtls,
)
from pypilecore.viewers.interactive_figures.figure_cpt_results_plan_view import (
    FigureCPTResultsPlanView,
)
from pypilecore.viewers.interactive_figures.figure_cpt_results_versus_ptls import (
    FigureCPTResultsVersusPtls,
)

__all__ = [
    "FigureCPTResultsVersusPtls",
    "FigureCPTResultsPlanView",
    "FigureCPTGroupResultsVersusPtls",
]



from __future__ import annotations  # noqa: F404

from typing import Any

import pandas as pd
from IPython.display import DisplayHandle, display
from ipywidgets import widgets
from natsort import natsorted

from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults
from pypilecore.viewers.interactive_figures import FigureCPTGroupResultsVersusPtls


class ViewerCptGroupResults:
    """
    Viewer for the CPT group results of the bearing capacity calculations.

    It offers the following layout:
        - Dropdown widgets:
            - Result: to select the result to show.
        - Figure CPT group results vs. pile tip level:
            - X axis: result values
            - Y axis: pile tip level w.r.t. NAP
            - Each trace represents a different case.
    """

    def __init__(self, cases_multi_results: CasesMultiCPTBearingResults) -> None:
        """Initialize the viewer.

        Parameters
        ----------
        cases_multi_results : CasesMultiCPTBearingResults
            The results of the bearing capacity calculations.

        Raises
        ------
        TypeError
            If 'cases_multi_results' are not of type 'CasesMultiCPTBearingResults'.
        """
        # Initialize figure CPT group resuls vs. pile tip level
        self._figure_plts = FigureCPTGroupResultsVersusPtls(
            cases_multi_results=cases_multi_results
        )

        # Set up control widgets
        _options = natsorted(
            pd.unique(cases_multi_results.cpt_group_results_dataframe.result_name)
        )
        self._result_dropdown = widgets.Dropdown(
            description="Result:",
            value=_options[0],
            options=_options,
        )

        # Update plot for initial selection of control widgets
        self._update_result(None)

        # Set up callbacks
        self._result_dropdown.observe(self._update_result, "value")

        # Set up layout
        self._control_widgets = widgets.HBox(
            [
                self._result_dropdown,
            ]
        )
        self._layout = widgets.VBox([self._control_widgets, self._figure_plts.figure])

    def _update_result(self, change: Any) -> None:
        """Private method to update the figure when the result name is changed in the result dropdown."""

        # Update the figure
        self._figure_plts.show_result(
            result_name=self._result_dropdown.value,
        )

    def display(self) -> DisplayHandle | None:
        """Display the figure."""
        return display(self._layout)



from typing import List, Protocol, TypeVar, runtime_checkable

from pypilecore.common.piles import PileProperties
from pypilecore.results.compression.multi_cpt_results import (
    CPTCompressionGroupResultsTable,
    SingleCPTCompressionBearingResultsContainer,
)
from pypilecore.results.tension.multi_cpt_results import (
    CPTTensionGroupResultsTable,
    SingleCPTTensionBearingResultsContainer,
)

T = TypeVar(
    "T",
    SingleCPTCompressionBearingResultsContainer,
    SingleCPTTensionBearingResultsContainer,
)
S = TypeVar("S", CPTCompressionGroupResultsTable, CPTTensionGroupResultsTable)


@runtime_checkable
class MultiCPTBearingResults(Protocol):
    """
    protocol classes for MultiCPTBearingResults from compression or tension endpoint response
    """

    def __init__(self) -> None:
        ...

    @property
    def pile_properties(self) -> PileProperties:
        ...

    @property
    def cpt_results(self) -> T:
        ...

    @property
    def cpt_names(self) -> List[str]:
        ...

    @property
    def group_results_table(self) -> S:
        ...
